MODULE   myFFTW0 ! non-mpi 1d fftw
use precision
implicit none
include  'fftw_f77.i'

type FFTW_type
  integer(PK) :: p, ip, rp, irp ! p: 'plan', pointer, 4 or 8 bytes, on 32- or 64-bit
end type
type(FFTW_type) :: FFTW_t
integer :: N
private :: FFTW_t, N

contains
!....................
subroutine init_fftw(Ni)
integer, intent(in) :: Ni
N = Ni
call fftw_f77_create_plan(FFTW_t%p,N, FFTW_FORWARD,  FFTW_estimate)
call fftw_f77_create_plan(FFTW_t%ip,N, FFTW_BACKWARD, FFTW_estimate)  
call rfftw_f77_create_plan(FFTW_t%rp,N, FFTW_REAL_TO_COMPLEX, FFTW_estimate)
call rfftw_f77_create_plan(FFTW_t%irp,N, FFTW_COMPLEX_TO_REAL, FFTW_estimate)
end subroutine
!....................
subroutine stop_fftw
call fftw_f77_destroy_plan(FFTW_t%p)
call fftw_f77_destroy_plan(FFTW_t%ip)
call rfftw_f77_destroy_plan(FFTW_t%rp)
call rfftw_f77_destroy_plan(FFTW_t%irp)
end subroutine
!....................
subroutine  fft1d(vin, vout)
COMPLEX, intent(in) :: vin(:)
COMPLEX, intent(out):: vout(:)
call fftw_f77_one(FFTW_t%p, vin, vout)
end  subroutine
!....................
subroutine ifft1d(vin, vout)
COMPLEX, intent(in) :: vin(:)
COMPLEX, intent(out):: vout(:)
call fftw_f77_one(FFTW_t%ip, vin, vout)
vout = vout / N
end  subroutine
!....................
subroutine rfft1d(vin, vout)  ! see p.7 of the fftw2.1.5 manual
! vout: [r0,r1,r2,...,r_{n/2}, i_{(n+1)/2-1},...,i2,i1]
real,intent(in)  :: vin(:)
real,intent(out) :: vout(:)
call rfftw_f77_one(FFTW_t%rp, vin, vout)
end  subroutine
!....................
subroutine irfft1d(vin, vout) ! <-- vin is destroyed!!!!!!!!
real, intent(inout)  :: vin(:)
real, intent(out)    :: vout(:)
call  rfftw_f77_one(FFTW_t%irp, vin, vout)
vout = vout/N
end  subroutine
!==============
end module
