MODULE   myMPI0
implicit none
include  'mpif.h'
integer  :: rank, nsize, ierr
! bare minimum of mpi

INTERFACE bcast  ! caution: to use these routines, all processors must 
  module procedure  bcast_r1, bcast_r2, bcast_i1, bcast_i2 ! have 
END INTERFACE    ! allocated the data in the same size.

contains
!-----------------------------------
subroutine init_mpi
call MPI_INIT(ierr)
call MPI_COMM_RANK(MPI_COMM_WORLD,rank, ierr)
call MPI_COMM_SIZE(MPI_COMM_WORLD,nsize,ierr)
end subroutine init_mpi
!-----------------------------------
subroutine stop_mpi
call MPI_Finalize(ierr)
end subroutine stop_mpi
!-----------------------------------
subroutine bcast_r1(v)
real,     intent(in) :: v(:)
call MPI_BCAST(v,size(v),MPI_REAL,  0,MPI_COMM_WORLD,ierr)
end  subroutine
!----------------------
subroutine bcast_r2(v)
real,      intent(in) :: v(:,:)
call MPI_BCAST(v,size(v),MPI_REAL,  0,MPI_COMM_WORLD,ierr)
end  subroutine
!----------------------
subroutine bcast_i1(v)
integer,   intent(in) :: v(:)
call MPI_BCAST(v,size(v),MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
end  subroutine
!----------------------
subroutine bcast_i2(v)
integer,   intent(in) :: v(:,:)
call MPI_BCAST(v,size(v),MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
end  subroutine
!-----------------------------------

subroutine partitionBegEnd(jbeg, jend, N, P, id, nbeg1, more2end1)
! N:     # of data points
! nbeg:  data points' index begins at nbeg (usu. 0, but could = PADDING, e.g.)
!        NBEG IS IN C CONVENTION!
! P:     # of parallel processors
! id:    a procerror's id, from 0 to P-1
!  We partition the data into P chunks, (default more2end == .false.)
!  If (more2end) then
!     remainders of N/P fill up from the last processor, 'coz the 0th 
!     is master processor, which has more work to do.
!  If (.not. more2end) then fill up from the 0th processor.
! 
! *jbeg  the beginning data index\__ for node i
! *jend  the ending data index   /
!        in fortran convention, index = 1,2,3,...
!        that is, if nbeg==0,   jbeg  = 1 (at id==0)
! 
! *jbeg > *jend  denotes empty data chunk assigned for this processor. 
integer, intent(out) :: jbeg, jend
integer, intent(in)  :: N, P, id
integer, intent(in), optional    :: nbeg1     ! [0]
LOGICAL, intent(in), optional    :: more2end1 ! [.false.]
integer  :: chnk, remn, nbeg
LOGICAL  :: more2end

nbeg    = 0;      if(present(nbeg1))     nbeg     = nbeg1
more2end=.false.; if(present(more2end1)) more2end = more2end1
chnk =  int(N / P);
remn =  N - P*chnk;
! if remn==0, then id'th node's data would range from 
!    id*chnk + 1 to (id+1)*chnk, inclusive.
jbeg =  id * chnk + nbeg + 1;! shift due to 'nbeg'
jend =  jbeg + chnk - 1;     ! index in Fortran concention
if (more2end) then
      ! Insert 1 datum for the last remn chunks: P-1, P-2, ..., P-remn
  if (id< P-remn)  return    ! unaffected
  chnk  = remn - (P-id)      ! now acts like a temp var
  jbeg  = jbeg + chnk
  jend  = jend + chnk + 1
else  ! Insert 1 datum for the 1st remn chunks: 0, 1, ..., remn-1
  if (id >= remn) then
    jbeg  = jbeg + remn
    jend  = jend + remn
  else
    jbeg  = jbeg + id
    jend  = jend + id + 1
  end if
end if
end subroutine
!=============
end module
