MODULE myOptim_mod

USE nrtype;  use myMath_module;  use myString_mod;  use seism_uti_module
use myMPI0
IMPLICIT NONE

INTERFACE polym_interp_steplength
module procedure polym_interp_steplength_s, polym_interp_steplength_d
END INTERFACE

real, pointer      :: x_cur(:,:), d_cur(:,:) ! current state and search
! dir. used in test func driver for line search
character(len=100) :: tmpfnm
PRIVATE :: x_cur, d_cur, tmpfnm
CONTAINS 
!==================================
subroutine Rosenbrock(x, f, g, J)
! acts as a test routine for optimization
real, intent(in) :: x(2)
real, optional, intent(out):: f, g(2), J(2,2)  ! J: Jacobian.  
! J(i,j) = dF(i)/dx(j), where F: 2 x 1 , and f = 0.5 F' F.
! J' J intends to approximate the Hessian like in Levenberg-Marquardt
if (present(f))   f = 50*(x(2) - x(1)*x(1))**2 + 0.5*(1 - x(1))**2
if (present(g))   then
  g(1) = 200* x(1)* (x(1)*x(1) - x(2)) + x(1) - 1
  g(2) = 100*(x(2) - x(1)*x(1))
endif
if (present(J))   then ! F = (/ 10*(x(2) - x(1).^2),  1 - x(1) /)
  J(1,:) = (/ -20.*x(1), 10. /)
  J(2,:) = (/ -1.,       0.  /)
endif
end subroutine
!.............
function  oscill_1d(x)  result(f) ! for testing 1d global optimization
real(SP), intent(in) :: x
real(SP)  :: f
f = x*sin(x) + x*cos(2.*x) ! g= sin(x)+ x cos(x)+cos(2 x)-2 x sin(2 x)
end function
!.............
subroutine cumsum2dLS(x, f, g, H) ! an exact least-square test function
real(SP), intent(in) :: x(:,:)
real(SP), optional, intent(out):: f, g(:,:), H !!! g is -gradient
! in presence of H, H = x' L' L x,   where x is serch direction, 
!    useful for analytical step length
integer   :: shp(2), j, k
real(SP), allocatable :: tmp1(:,:), tmp2(:,:)
shp = shape(x)
! L operator is 2d_cum_sum of x
! cum_sum(dim1) and cum_sum(dim2) commute
! cum_sum(dim1)'s operator is lower triangle matrix of 1's;  so,
! L' is upper triang of 1's, i.e., cum_sum from right to left.
! inv(L) is diff0()
allocate ( tmp1(shp(1),shp(2)), tmp2(shp(1),shp(2)) )
call cumsum2d(tmp1, x,    1)
call cumsum2d(tmp2, tmp1, 2)   ! tmp2 = L x
if(present(H)) then
  H = norm2(tmp2)
end if
tmp2 = tmp2 - 1  ! 1 being the target array
if(present(f)) then
  f = norm2(tmp2)*0.5 ! minimal f=0 is achieved when x = delta(1,1)
end if
if(present(g)) then   ! 
  call cumsum2d(tmp1, tmp2(shp(1):1:-1, shp(2):1:-1), 2)
  call cumsum2d(g, tmp1, 1)
  g = -g(shp(1):1:-1, shp(2):1:-1) 
end if
deallocate (tmp1, tmp2)
end subroutine
!.............
function cumsum2dLS_fun(alph) result(fun) ! driver for line search
real, intent(inout) :: alph
real  ::  fun
call  cumsum2dLS(x_cur + alph*d_cur, fun) 
end function
!............. 

subroutine cumsum2dLS_minimize ! test routine
! start from 0
integer,      parameter   :: Mx = 2, Nx = 3
real, target, allocatable :: mem(:,:,:)
real, pointer :: p0(:,:), p1(:,:), r0(:,:), r1(:,:), z1(:,:)
! see CG_precon_reset() for comments
real          :: misf(Mx*Nx+1, 4) 
! ^ Stp Desc;  CG;  CG w some precond;  CG w perfect precond
integer       :: k

!call init_mpi
allocate(mem(Mx, Nx, 6))
x_cur => mem(:,:,1)
p0    => mem(:,:,2);    p1    => mem(:,:,3)
r0    => mem(:,:,4);    r1    => mem(:,:,5)
z1    => mem(:,:,6)
do k = 4, 4
  call main_iter(k, misf(:,k))
  call filename(tmpfnm, 'State2x3_mode', k, '.dat')
  call write_binfile(tmpfnm, x_cur)
end do
call write_binfile('cnvrg_crvs7x4.bin', misf)
deallocate(mem)
!call stop_mpi
 
contains

  subroutine main_iter(mode, msft)
  integer, intent(in) :: mode ! 1, 2, 3, 4:  
  ! {StpDesc, CG, CGsomePrecn, CG perfectPrecn}
  real,    intent(out):: msft(:)
  integer  :: it, RESET
  real     :: xmin, stpsz, stpAna
  RESET = 10000;  if (mode==1) RESET = 1

  x_cur = 0. 
  msft  = 0.  ! init.
  call cumsum2dLS(x_cur, msft(1), r1)  ! for it==0
  select case (mode)
    case (1:2)
      z1 = r1
    case (3)
      call precond_half(z1, r1)
    case (4)
      call precond(z1, r1)
  end select
  call write_binfile('prcnd_grd.bin', z1)
  do it = 1, Mx*Nx
    call CG_precon_reset(p0,p1, r0,r1, z1, RESET)
    d_cur => p0
    stpsz = 0.5
    msft(it+1)= optimz_1d(xmin, 0., stpsz, msft(it), cumsum2dLS_fun, 5e-3)
    call cumsum2dLS(d_cur, H=stpAna)
    stpAna= dot_prod(r0, z1)/stpAna
    write(*,131)  it, xmin, stpAna; call flush(6)
131 format ('*** it: numer, analy  Step', 3I, 2ES11.3)   
    if(it==1) call scan_a_line(9, xmin, cumsum2dLS_fun, 'J_line.bin')
    
    x_cur = x_cur + xmin * d_cur
    if (msft(it+1) < 1e-3)   exit
    
    call cumsum2dLS(x_cur, g=r1)
    select case (mode)
      case (1:2)
        z1 = r1
      case (3)
        call precond_half(z1, r1)
      case (4)
        call precond(z1, r1)
    end select
  end do
  end subroutine

  subroutine precond(z,r)  ! exact inv(L'L) r = inv(L) * inv(L') r
  real, intent(out) :: z(:,:)
  real, intent(in)  :: r(:,:)
  z = diff0(r(Mx:1:-1, Nx:1:-1),1)
  z = diff0(z, 2)
  z = z(Mx:1:-1, Nx:1:-1)  ! done inv(L') r
  z = diff0(z,1)
  z = diff0(z,2)
  end subroutine

  subroutine precond_half(z,r)  ! half way precondi
  real, intent(out) :: z(:,:)
  real, intent(in)  :: r(:,:)
  call precond(z,r)
  z = 0.5*(z+r)
  end subroutine

end subroutine



!++++++++++++++++++++++++++++
subroutine scan_a_line(Nstps, stp0, func, fname)
! probe the criterion function along a line
integer, intent(in) :: Nstps ! # of steps to scan, including 0
real,    intent(in) :: stp0  ! estimated best step to arrive at minimum
INTERFACE
 FUNCTION  func(x)
 USE nrtype
 REAL(SP), INTENT(inout) :: x
 REAL(SP)  :: func
 END FUNCTION
END INTERFACE
character(*), intent(in) :: fname ! to save in this file
real,allocatable    :: msft(:,:)
real    :: stp, stpk
integer :: k
allocate(msft(Nstps,2))
stp  = 2.*stp0/(Nstps - 1)
do k = 0, Nstps-1
  stpk = k* stp
  msft(k+1,1) = stpk
  msft(k+1,2) = func(stpk)
end do
call write_binfile(fname, msft)
deallocate(msft)
end subroutine
!++++++++++++++++++++++++++++



subroutine lineSearch_backtracking_0(lam, f1, f0, fp0, func)
! line search, polynomial interpolating and Armijo condition
!     '0':  the mother version.
!     (current point & gradient are embedded in func)
! lam=0 <--> current point
! input  lam: 1st guess of step size
! output lam: next acceptable point.  f1 = func(final_lam)
! f0 = func(0)
! fp0: d func / d lam, the slope @ lam=0, must be < 0, i.e., downhill

real, intent(inout) :: lam
real, intent(out)   :: f1
real, intent(in)    :: f0, fp0
INTERFACE
FUNCTION func(x)
REAL, INTENT(IN):: x  ! scalar, step length
REAL  :: func
END FUNCTION
END INTERFACE

real    :: lam_tmp, lam_prev, f_prev, c1 = 1e-4
integer :: Nvio     ! # times Armijo cond. is violated
Nvio = 0
do
  f1 = func(lam)
  if (f1 < f0 + c1*lam*fp0)  return
  Nvio= Nvio + 1;   ! write(*,*) 'Nvio', Nvio
  if (Nvio == 1) then ! quadratic fit
    lam_tmp = polym_interp_steplength(f0, fp0, lam, f1)
  else
    lam_tmp = polym_interp_steplength(f0, fp0, lam, f1, lam_prev, f_prev)
  end if
  lam_prev = lam;  f_prev = f1
  lam = lam_tmp
end do
end subroutine
!-------------
function polym_interp_steplength_s(f0,fp0,alpc,fc,alpl,fl) result(alph)
! single precision
! Cubic/quadratic linesearch step length
! Finds minimizer 'alph' of the cubic polynomial f  within the interval
!      [0.10, 0.51] * alpc, such that
! f(0) =f0,  f'(0) =fp0,  f(alpc) =fc,  f(alpl) =fl, where suffixes are
!     ...c  current
!     ...l  last
!     ...h  hat, the one in question.
! If data for a cubic is not available (1st stepsize reduction) then f
!    is the quadratic such that
! f(0) = f0,  f'(0) = fp0,  f(alpc) = fc
! Ref:
! [1] Nocedal & Wright, "Numerical Optimization", 2nd Ed., Eq.3.58
real, optional    :: alpl, fl
real, intent(in)  :: f0, fp0, alpc, fc
real  :: alph,  det, dl, dc, a, b
if (.not.present(fl)) then ! quadratic model
  alph= -fp0 *alpc*alpc /(2*(fc - f0 - fp0*alpc))
else             ! cubic model
  det = alpc*alpc*alpl*alpl*(alpl - alpc)
  dl  = fl - f0 - fp0*alpl
  dc  = fc - f0 - fp0*alpc
  a   = (alpc*alpc *dl - alpl*alpl *dc)/det
  b   = (-alpc**3  *dl + alpl**3   *dc)/det
  alph= (sqrt(b*b - 3*a*fp0) - b)/3/a
endif
if (alph < 0.1*alpc)  then
  alph =  0.1 *alpc
else if (alph > 0.51*alpc) then
  alph =  0.51*alpc
endif
end function

function polym_interp_steplength_d(f0,fp0,alpc,fc,alpl,fl) result(alph)
! Cubic/quadratic linesearch step length  in  double precision
real(DP), optional    :: alpl, fl
real(DP), intent(in)  :: f0, fp0, alpc, fc
real(DP)  :: alph,  det, dl, dc, a, b
if (.not.present(fl)) then ! quadratic model
  alph= -fp0 *alpc*alpc /(2*(fc - f0 - fp0*alpc))
else             ! cubic model
  det = alpc*alpc*alpl*alpl*(alpl - alpc)
  dl  = fl - f0 - fp0*alpl
  dc  = fc - f0 - fp0*alpc
  a   = (alpc*alpc *dl - alpl*alpl *dc)/det
  b   = (-alpc**3  *dl + alpl**3   *dc)/det
  alph= (sqrt(b*b - 3*a*fp0) - b)/3/a
endif
if (alph < 0.1*alpc)  then
  alph =  0.1 *alpc
else if (alph > 0.51*alpc) then
  alph =  0.51*alpc
endif
end function

!sssssssssssssssssssssssssssssssssssssssssssssssssssssssssssss
subroutine CG_precon_reset(p0,p1, r0,r1,z1, RESETIT, Nreset, hit_n_reset, rk)
!          One iterarion step of C.G.
! input:   r1,z1, RESETIT, Nreset,   rk
! output:  p0 (current CG search dir.)  --> Bcast to all proc. in the end,
!          hit_n_reset.
! for non-root processes, only receive p0, no need to provide input arrays.
! p:   CG search direction (use Polak-Ribiere rule in computing beta)
! r:   negative gradient
! z:   preconditioned r
! subscripts' meaning:     p0/p1: current/next;      r0/r1: old/new
!
! RESETIT:  reset to steepest descent every RESETIT iter.'s
! Nreset:   # of resets, once reached, return  hit_n_reset = .true.
! rk:       rank of the process, its presence denoting MPI

! This routine is typically embedded in an iterative scheme, who is 
! responsible for maintaining the memory for p0,p1, r0,r1, z1.  So,
!   ***** NEVER WRITE ON  p0, p1, r0, *****
!   which are persistent data required by this routine.
!
! As pointer swapping is more efficient than memory copying, this
! routine does not maintain persistent memory space on its own;
! instead it expects the calling party to provide the necessary
! memory space.
! Recall in FORTRAN, pointer is always passed by reference, i.e.,
! like intent(inout), therefore swapping pointers in this routine
! == swapping the calling party's pointers.
!
! Implementation detail:  
! p1 gets the new result first, but p0 & p1 are then swapped.  
! So p0 has the current dir. in the end.  At the beginning of next
! call, p0 acts as the old serch direction.

real,    pointer    :: p0(:,:), p1(:,:), r0(:,:), r1(:,:), z1(:,:)
integer, intent(in) :: RESETIT
logical, intent(out),  optional :: hit_n_reset
integer, intent(in),   optional :: Nreset,        rk

real                :: dot1, dotc, beta
real,    SAVE       :: dot0
integer, SAVE       :: it = 0,     n_reset = 0
integer             :: rank, numel ! # elements in r,z,p

 rank  = 0;         if(present(rk))        rank   = rk
 if(present(Nreset) .neqv. present(hit_n_reset))  &
    stop 'Nreset & hit_n_reset must apper together, in CG_precon_reset().'
	
 if (RESETIT == 1)  then  ! simply steepest descn
   if (rank==0)     then
      p1 = z1     ! search dir. is just the preconditioned  -grad
      call swap_pointers(r0, r1)
      call swap_pointers(p0, p1)
   end if
   numel = size(p0)
   if(present(rk))  call MPI_BCAST(p0, numel, MPI_REAL, 0, &
                         MPI_COMM_WORLD, ierr)
   if(present(Nreset))   then					
	    n_reset = mod(n_reset + 1,  Nreset)
	    hit_n_reset = n_reset ==0
   end if
   return
 end if
 
 if (rank==0) then
    dot1 = dot_prod(z1, r1)
    if (it==0) then
       p1   = z1  ! set to steepest descn
    else
       dotc = dot_prod(z1, r0)
       beta = (dot1 - dotc)/dot0
       p1   = z1 + beta*p0
#ifdef DEBUGo
       write(*,*) '<r0,z1>, <p0,r1>', dotc, dot_prod(p0, r1) ! should==0
#endif
    end if

    call swap_pointers(r0, r1)
    call swap_pointers(p0, p1)
    dot0 = dot1
    it   = mod(it+1, RESETIT)
    if (it==0 .and. present(Nreset))   then
       n_reset = mod(n_reset + 1,  Nreset)
	     hit_n_reset = n_reset ==0
    else
       hit_n_reset = .false.
    end if
 end if
 numel = size(p0)
 if(present(rk)) call MPI_BCAST(p0, numel, MPI_REAL, 0, MPI_COMM_WORLD, ierr)
end subroutine

!sssssssssssssssssssssssssssssssssssssssssssssssssssssssssssss
SUBROUTINE dmnbrak(ax,bx,cx, fa,fb,fc, func, rank, Nfeval, rec)
! in MPI, only root has valid input/output results;
!         func() needs all proc. to participate
! what's different from the NR routine mnbrak() is:
! here we know it's downhill at ax toward bx.  So, instead of swapping a & b as in
! NR, we can try reducing the interval width to get a nadir.

! USE nrutil, ONLY : swap
REAL(SP), INTENT(INOUT) :: ax, bx, fa  ! as fa is often known after ax is arrived at
! by earlier iter., so, don't compute fa again, instead, consider it available!
REAL(SP), INTENT(OUT)   :: cx, fb, fc
integer(I4B), optional, INTENT(OUT):: Nfeval   ! # func. evalu.
REAL(SP),     optional, INTENT(OUT):: rec(:,:) ! record [u, fu; ...] Nfeval X 2
                             ! the presence of rec demands the presence of Nfeval
integer(I4B), optional, INTENT(IN) :: rank     ! MPI, calling processor's rank.
! if rank is absent or rank<0 ==> no MPI

! func(x) requires all processors participate, but only root handles bracketing
! All MPI proc must follow exactly the same path forking at conditionals, otherwise
! MPI_BCAST will hang, expecting everyone to participate.  So, at conditionals,
! root must tell others (SIGNAL) what subpath it takes.  At a conditional, for every
! possibility, root must BCAST to others the SIGNAL.  (Otherwise if in some possibility
! root doesn't BCAST, yet others are waiting for the SIGNAL, then MPI hangs.)

! here, the x in others' func(x) is just dummy.  In func(x), however, root should
! BCAST to others the value of x.  After collaboration, the final function value is
! reduced to root.  Again, others don't need to have that value, because only the root
! handles optimization decisions.

INTERFACE
 FUNCTION func(x)
  USE nrtype
  REAL(SP),INTENT(inout):: x
  REAL(SP)           ::  func
 END FUNCTION
END INTERFACE

REAL(SP),  PARAMETER :: GOLD=2.618034_sp, GLIMIT=100.0_sp, TINY=1.0e-20_sp ! GOLD=1.618034_sp
           ! I changed Numer. Rec.'s GOLD to GOLD^2, so that increasing bracket search is faster
REAL(SP)     :: fu, q,r, u,ulim
INTEGER(I4B) :: rnk,  SIGNAL
logical      :: useMPI

rnk = 0;     useMPI = .false.
if(present(rank)) then
  if (rank>=0)    then
     rnk = rank;  useMPI = .true.
  end if
end if
u   = 0.  ! dummy init for non-root

! fa   already available
fb  = func(bx)
if (rnk==0) then  ! ##############
    if (present(Nfeval)) Nfeval = 0;  call debug_show('b, fb', bx, fb)
  if (fb > fa) then
    ! call swap(ax,bx); call swap(fa,fb)   ! old NR code
    SIGNAL = 1
    if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    DO   ! my modification follows, in bold
      U = AX + (BX-AX)/4.    ! shrink at 1/4th
      FU= FUNC(U);  call debug_show('1/4 shrink: u, fu:', u, fu)
      IF (FU >= FA) THEN
        BX = U;  FB = FU
        SIGNAL = 3
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      ELSE
        CX = BX; FC = FB
        BX = U;  FB = FU
        SIGNAL = 4
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        RETURN
      END IF
    END DO
  else
    SIGNAL = 0
    if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
  end if
else  ! ############## other proc's closely follow root's path
  if (present(Nfeval))  Nfeval = 1
  call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
  if (SIGNAL==1) then
    DO
      FU= FUNC(U);  if (present(Nfeval))  Nfeval=Nfeval+1
      call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      if (SIGNAL==4)  RETURN
    END DO
  end if
end if
 cx = bx + GOLD*(bx-ax)
 fc = func(cx)
if (rnk==0) then 
  call debug_show('gold 215: c, fc:', cx, fc)
else
  if (present(Nfeval))  Nfeval=Nfeval+1
end if
do
  if (rnk==0) then  ! ##############
    if (fb < fc) then
      SIGNAL = 1
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      RETURN
    else
      SIGNAL = 0
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    end if
    r= (bx-ax)*(fb-fc)
    q= (bx-cx)*(fb-fa)
    u= bx-((bx-cx)*q-(bx-ax)*r)/(2.0_sp*sign(max(abs(q-r),TINY),q-r))
    ulim=bx + GLIMIT*(cx-bx)
    if ((bx-u)*(u-cx) > 0.0) then
      SIGNAL = 2   !@@@@@@@
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      fu= func(u)
        call debug_show('parabola 237: u, fu:', u, fu)
      if (fu < fc) then
        SIGNAL = 3     !$$$$$$
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        ax =bx;  fa = fb
        bx = u;  fb = fu
        RETURN
      else if (fu > fb) then
        SIGNAL = 4     !$$$$$$
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        cx = u
        fc =fu
        RETURN
      else
        SIGNAL = 40    !$$$$$$
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      end if
      u = cx+GOLD*(cx-bx)
      fu= func(u);  call debug_show('replace parabo w/ gold 255: u, fu:', u, fu)
    else if ((cx-u)*(u-ulim) > 0.0) then
      SIGNAL = 5    !@@@@@@@
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      fu= func(u); call debug_show('Parabola 259: u, fu:', u, fu)
      if (fu < fc) then
        SIGNAL = 6         !$$$$$$
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        bx= cx
        cx= u
        u = cx+ GOLD*(cx-bx)
        call shft(fb,fc,fu,func(u));  call debug_show('gold 266: u, fu:', u, fu)
      else
        SIGNAL = 60         !$$$$$$
        if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      end if
    else if ((u-ulim)*(ulim-cx) >= 0.0) then
      SIGNAL = 7    !@@@@@@@
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      u = ulim
      fu= func(u);  call debug_show('clip: u, fu:', u, fu)
    else
      SIGNAL = 8    !@@@@@@@
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      u = cx + GOLD*(cx-bx)
      fu= func(u);  call debug_show('gold 280: u, fu:', u, fu)
    end if
    call shft(ax,bx,cx, u)
    call shft(fa,fb,fc,fu)
  else ! ##############  other proc's closely follow root's path
    call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    if (SIGNAL == 1) return
    
    call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    if(SIGNAL == 2) then        !@@@@@@@
      fu= func(u);  if (present(Nfeval))  Nfeval=Nfeval+1
      call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      if (SIGNAL==3 .or. SIGNAL==4) then    !$$$$$$
        return      
      end if
      fu= func(u);  if (present(Nfeval))  Nfeval=Nfeval+1
    else if (SIGNAL == 5) then  !@@@@@@@
      fu= func(u);  if (present(Nfeval))  Nfeval=Nfeval+1
      call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      if (SIGNAL == 6)  then
        fu = func(u)      !$$$$$$
        if (present(Nfeval))  Nfeval=Nfeval+1
      end if
    else if (SIGNAL == 7) then  !@@@@@@@
      fu= func(u);  if (present(Nfeval))  Nfeval=Nfeval+1 
    else  ! has to be 8         !@@@@@@@
      fu= func(u);  if (present(Nfeval))  Nfeval=Nfeval+1
    end if
    
  end if ! ##############
end do

CONTAINS  !==============
! private subroutine:
  SUBROUTINE shft(a,b,c,d)
  REAL(SP), INTENT(OUT)   :: a
  REAL(SP), INTENT(INOUT) :: b, c
  REAL(SP), INTENT(IN)    :: d
  a=b;  b=c;  c=d
  END SUBROUTINE shft

  SUBROUTINE debug_show(mssg, xx, fxx)
  character(len=*),intent(in) :: mssg
  real, intent(in) :: xx, fxx
  if (present(Nfeval))  Nfeval=Nfeval+1
#ifdef DEBUGo
  write(*,*) mssg, xx, fxx
  if (present(Nfeval)) write(*,*) '...@ Nfeval', Nfeval
  call flush(6)
#endif 
  if (present(rec))     then
    if (Nfeval> size(rec,1)) then
      write(*,*) 'Nfeval, shape(rec)', Nfeval, shape(rec)
      stop 'dmnbrak: size(rec,1) < Nfeval!!!'
    end if  
    rec(Nfeval,1) =  xx 
    rec(Nfeval,2) = fxx
  end if  
  END SUBROUTINE debug_show

END SUBROUTINE dmnbrak



FUNCTION my_brent(xmin, ax,bx,cx, fa,fb,fc, func, tol, rank, Nfeval, rec)
! in MPI, only root has valid input/output results;
!         func() needs all proc. to participate
! what's different from the NR brent() is:
! from bracketing, we know the function values at a,b,c.  So we don't need to
! calculate them again.  Initialize w & v to a & c; this makes parabola fit
! immediately possible.  In contrast, in NR, w & v & x are initialized to the
! same point b;  this means we have to make 2 Golden steps before parabola fit
! is possible.  That is, the start up speed suffers in NR's brent()
REAL(SP),     INTENT(OUT)    :: xmin
REAL(SP),     INTENT(IN)     :: ax,bx,cx, fa,fb,fc, tol
REAL(SP)      :: my_brent
integer(I4B), optional, INTENT(OUT):: Nfeval   ! # func. evalu.
REAL(SP),     optional, INTENT(OUT):: rec(:,:) ! record [u, fu; ...] Nfeval X 2
                             ! the presence of rec demands the presence of Nfeval
integer(I4B), optional, INTENT(IN) :: rank     ! MPI, calling processor's rank
! if rank is absent or rank<0 ==> no MPI

! func(x) requires all processors participate, but only root handles bracketing
! See the comment in dmnbrak()

INTERFACE
 FUNCTION  func(x)
 USE nrtype
 REAL(SP), INTENT(inout) :: x
 REAL(SP)             :: func
 END FUNCTION
END INTERFACE

INTEGER(I4B),PARAMETER:: ITMAX = 3 !!!!!!!!!
REAL(SP),    PARAMETER:: CGOLD=0.3819660_sp, ZEPS=1.0e-3_sp*epsilon(ax)
INTEGER(I4B) :: iter,  rnk,  SIGNAL
logical      :: useMPI
REAL(SP)     :: a,b,d,e,etemp, fu,fv,fw,fx, p,q,r, tol1,tol2, u,v,w,x,xm

rnk = 0;     useMPI = .false.
if(present(rank)) then
  if (rank>=0)    then
     rnk = rank;  useMPI = .true.
  end if
end if
if (present(Nfeval))  Nfeval = 0     ! init 

if (rnk==0) then
  x = bx;  fx = fb
  if (fa < fc) then
    w = ax;  fw = fa;
    v = cx;  fv = fc;
  else
    w = cx;  fw = fc;
    v = ax;  fv = fa;
  end if
  a = min(ax,cx)
  b = max(ax,cx)
  d = 1.0e10   ! always allow first step to use parabola fit
  e = d
#ifdef DEBUGo
  write(*,628) ax,bx,cx, fa,fb,fc;   call flush(6)
628 format(/ 'Brent: ax,bx,cx, fa,fb,fc', 6ES11.3)
#endif
end if

u = 0.
do iter= 1,ITMAX
  if (rnk==0) then ! MPI root, or no MPI #######################
    xm   = 0.5_sp*(a+b)
    tol1 = tol*abs(x)+ZEPS
    tol2 = 2.0_sp*tol1
    if (abs(x-xm) <= (tol2-0.5_sp*(b-a))) then
#ifdef DEBUGo
      write(*,642) ax,bx,cx, fa,fb,fc
      write(*,643) a, xm, x, b, tol2;  call flush(6)
642 format('ax,bx,cx, fa,fb,fc', 6ES11.3)
643 format('a, xm, x, b, tol2',  5ES11.3)
#endif
      xmin    = x
      my_brent= fx
      SIGNAL = 1
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
      RETURN
    else
      SIGNAL = 0
      if (useMPI) call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    end if
    if (abs(e) > tol1) then
      r=(x-w)*(fx-fv)
      q=(x-v)*(fx-fw)
      p=(x-v)*q-(x-w)*r
      q=2.0_sp*(q-r)
      if (q > 0.0) p = -p
      q=abs(q)
      etemp=e
      e    =d
      if (abs(p) >= abs(0.5_sp*q*etemp) .or. &
               p <= q*(a-x) .or. p >= q*(b-x)) then
        e=merge(a-x, b-x, x >= xm )
        d=CGOLD*e
#ifdef DEBUGo
        write(*,*) 'GOLD'; call flush(6)
#endif        
      else
        d= p/q
        u= x+d
        if (u-a < tol2 .or. b-u < tol2) d=sign(tol1,xm-x)
#ifdef DEBUGo
        write(*,*) 'PARAB'; call flush(6)
#endif             
      end if
    else
      e= merge(a-x, b-x, x >= xm )
      d= CGOLD*e
#ifdef DEBUGo 
        write(*,*) 'GOLD'; call flush(6)
#endif        
    end if
#ifdef DEBUGo
    write(*,*) 'd, tol1, x', d, tol1, x; call flush(6)
#endif     
    u = merge(x+d, x+sign(tol1,d), abs(d) >= tol1)
    fu= func(u);   call debug_show('Brent: u, fu', u, fu)  !!!!!!!!!
    if (fu <= fx)  then
      if (u >= x)  then
        a=x
      else
        b=x
      end if
      call shft(v,w,x,u)
      call shft(fv,fw,fx,fu)
    else
      if (u < x) then
        a=u
      else
        b=u
      end if
      if (fu <= fw .or. w == x) then
        v = w
        fv=fw
        w = u
        fw=fu
      else if (fu <= fv .or. v == x .or. v == w) then
        v = u
        fv=fu
      end if
    end if
#ifdef DEBUGo
    write(*,716) a,v,w,x,b,  fv,fw,fx;    call flush(6)
716 format('a,v,w,x,b', 5ES11.3, / 'fv,fw,fx', 3ES11.3)
#endif
  else  ! non-roots, rnk/=0 could happen only if useMPI
        ! #######################
    call MPI_BCAST(SIGNAL, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    if (SIGNAL == 1)   RETURN
    fu= func(u) ! this u is dummy.  In func, true u is BCAST'ed from root to others
    if (present(Nfeval))  Nfeval=Nfeval+1
  end if! #######################
end do  ! iter
xmin = x;   my_brent= fx

CONTAINS ! private subroutine:
  SUBROUTINE shft(a,b,c,d)
  REAL(SP), INTENT(OUT)   :: a
  REAL(SP), INTENT(INOUT) :: b, c
  REAL(SP), INTENT(IN)    :: d
  a=b;  b=c;  c=d
  END SUBROUTINE shft
  
  SUBROUTINE debug_show(mssg, xx, fxx)
  character(len=*),intent(in) :: mssg
  real, intent(in) :: xx, fxx
  if (present(Nfeval))  Nfeval = Nfeval + 1
#ifdef DEBUGo
  write(*,*) mssg, xx, fxx
  if (present(Nfeval))  write(*,*) '...@ Nfeval', Nfeval
  call flush(6)
#endif 
  if (present(rec))     then
    if (Nfeval> size(rec,1)) then
      write(*,*) 'Nfeval, shape(rec)', Nfeval, shape(rec)
      stop 'dbrent: size(rec,1) < Nfeval!!!'
    end if  
    rec(Nfeval,1) = xx; rec(Nfeval,2) = fxx
  end if 
  END SUBROUTINE debug_show
END FUNCTION my_brent


FUNCTION optimz_1d(xmin, ax0,bx0,fa0, func, tol, ranki, Nfevali, reci) result(fmin)
! in MPI, only root has valid input/output results;
!                                only func() needs all proc. to participate
! combine my bracketing and brent
REAL(SP),       INTENT(OUT)    :: xmin
REAL(SP),       INTENT(IN)     :: ax0, bx0, fa0, tol
REAL(SP)        :: fmin
INTERFACE
 FUNCTION func(x)
 USE nrtype
 REAL(SP),INTENT(inout) :: x
 REAL(SP)            :: func
 END FUNCTION
END INTERFACE
integer(I4B), optional, INTENT(OUT):: Nfevali ! # func. evalu., if present --> debug mode
REAL(SP),     optional, INTENT(OUT):: reci(:,:) ! record [u, fu; ...] Nfeval X 2
                             ! the presence of rec demands the presence of Nfeval
integer(I4B), optional, INTENT(IN) :: ranki     ! MPI, calling processor's rank
! xxxi:  'i' is to avoid same name w/ optional argument of called routines

integer(I4B)            :: N1, N2, rnk
REAL(SP)                :: ax, bx, cx, fa, fb, fc

rnk = -1;  if(present(ranki))  rnk = ranki      ! '-1' denotes no MPI
ax = ax0;   bx = bx0;   fa = fa0

if (present(reci)) then
	call dmnbrak(ax,bx,cx, fa,fb,fc, func, rnk, N1, reci)
	fmin = my_brent(xmin, ax,bx,cx, fa,fb,fc, func, tol, rnk, N2, reci(N1+1:,:))
	Nfevali = N1 + N2
	call debug_show

else if (present(Nfevali)) then
	call dmnbrak(ax,bx,cx, fa,fb,fc, func, rnk, N1)
	fmin = my_brent(xmin, ax,bx,cx, fa,fb,fc, func, tol, rnk, N2)
	Nfevali = N1 + N2
	call debug_show

else
	call dmnbrak(ax,bx,cx, fa,fb,fc, func, rnk)
	fmin = my_brent(xmin, ax,bx,cx, fa,fb,fc, func, tol, rnk)
end if

contains

 subroutine debug_show
 if(rnk <= 0)  write(*,999) N1,N2, bx0, xmin, xmin/bx0
 call flush(6)
999 format ('N1,N2, Stp_ini,Stp_fin, ratio', 2I3, 3ES12.3) 
 end subroutine
 
END FUNCTION

END MODULE

