! utility subroutines for reading input parameters
! and for parsing intergers, etc.
MODULE myPARSER

implicit none
integer,  private,   parameter  ::  cdim = 200, nline= 250
INTERFACE READCMD
  module  procedure  readCmdString,   readCmdInteger,  readCmdReal, &
                     readCmdInteger1, readCmdLogical
end interface

interface Str2Vector_VaryingLen
  module  procedure  Str2IntVector_VaryingLen, Str2RealVector_VaryingLen
end interface

INTERFACE READPARFILE
  module procedure   readParFileString,  readParFileInteger, readParFileReal, &
                     readParFileLogical, readParFileInteVec, readParFileRealVec, &
                     readParFileLogical1,readParFileInteger1
end interface

CONTAINS
!=================================================================
function  extract_digit(a, d, bas)
! extract the dth digit of an integer a of base bas, default 10
integer   ::  extract_digit,  tmp, b
integer,  intent(in) :: a, d
integer,  intent(in), optional :: bas
if (.not. present(bas)) then
  b = 10
else
  b = bas
endif
tmp = a/(b**(d-1))
extract_digit = tmp - (tmp/b)*b
end function
!--------------------
subroutine Str2IntVector_VaryingLen(str, ivec, nelem)
character(*), intent(in)  :: str
integer,      intent(out) :: ivec(:), nelem
! str:  e.g., 6,1,34  or 9, 21  3, i.e., ',' & ' ' are treated as separators
! ivec: vector of those integers
! nelem: # of extracted elements, make sure ivec is big enough
integer :: len_str, len_ivec, head, tail, loc  ! head, tail: of the last integer in str
len_str = len_trim(str);      len_ivec = size(ivec)
nelem   = 0;  tail = 0; 
do
  loc  = verify(str(tail+1:),', ') ! find the 1st char not a ',' or ' '
  if (loc==0)  exit ! no valid numbers left
  nelem= nelem + 1
  if (nelem > len_ivec) stop '# integer elements in the string > size of integer vector!'
  head = tail + loc
  loc  = scan(str(head+1:), ', ')
  if (loc==0) then  ! last integer element in str, no trailing ' ' or ','
    tail = len_str
  else 
    tail = head + loc - 1
  end if
  read(str(head:tail),*)  ivec(nelem)
end do
end subroutine
!--------------------
subroutine Str2RealVector_VaryingLen(str, rvec, nelem)
character(*), intent(in)  :: str
real,         intent(out) :: rvec(:)
integer,      intent(out) :: nelem
! see Str2IntVector_VaryingLen()
integer :: len_str, len_rvec, head, tail, loc  ! head, tail: of the last integer in str
len_str = len_trim(str);      len_rvec = size(rvec)
nelem   = 0;  tail = 0; 
do
  loc  = verify(str(tail+1:),', ') ! find the 1st char not a ',' or ' '
  if (loc==0)  exit ! no valid numbers left
  nelem= nelem + 1
  if (nelem > len_rvec) stop '# real elements in the string > size of real vector!'
  head = tail + loc
  loc  = scan(str(head+1:), ', ')
  if (loc==0) then  ! last real element in str, no trailing ' ' or ','
    tail = len_str
  else 
    tail = head + loc - 1
  end if
  read(str(head:tail),*)  rvec(nelem)
end do
end subroutine

!--------------------
subroutine readCmdString(keyword, cvalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  character(*), intent(in), optional :: default_val
  character(*), intent(in)           :: keyword
  character(*), intent(out)          :: cvalue
  integer           :: iargc, i
  logical           :: ismissing
  character(cdim)   :: argument

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  if (iargc() == 0) then   ! There is no input argument.
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!'; stop
    endif
    if (present(default_val)) cvalue = default_val   ! Use the default_val value.
  else
    do i = 1, iargc()
      call getarg(i, argument)
      if (argument(1:len(keyword)) == keyword) then
        cvalue = argument(len(keyword)+2:len_trim(argument))
        return
      endif
    enddo
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!';  stop
    endif
    if (present(default_val))  cvalue = default_val   ! Use the default_val value.
  endif
end subroutine 

!--------------------------------------------------------------------
subroutine readCmdInteger1(keyword, ivalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  integer(1),   intent(in), optional :: default_val
  character(*), intent(in)           :: keyword
  integer(1),   intent(out)          :: ivalue
  integer           :: iargc, i
  logical           :: ismissing
  character(cdim)   :: argument

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  if (iargc() == 0) then   ! There is no input argument.
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!'; stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  else
    do i = 1, iargc()
      call getarg(i, argument)
      if (argument(1:len(keyword)) == keyword) then
        read(argument(len(keyword)+2:len_trim(argument)),*) ivalue
        return
      endif
    enddo
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!';  stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  endif
end subroutine

!--------------------------------------------------------------------
subroutine readCmdInteger(keyword, ivalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  integer,      intent(in), optional :: default_val
  character(*), intent(in)           :: keyword
  integer,      intent(out)          :: ivalue
  integer           :: iargc, i
  logical           :: ismissing
  character(cdim)   :: argument

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  if (iargc() == 0) then   ! There is no input argument.
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!'; stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  else
    do i = 1, iargc()
      call getarg(i, argument)
      if (argument(1:len(keyword)) == keyword) then
        read(argument(len(keyword)+2:len_trim(argument)),*) ivalue
        return
      endif
    enddo
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!';  stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  endif
end subroutine

!--------------------------------------------------------------------
subroutine readCmdLogical(keyword, ivalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  logical,      intent(in), optional :: default_val
  character(*), intent(in)           :: keyword
  logical,      intent(out)          :: ivalue
  integer           :: iargc, i
  logical           :: ismissing
  character(cdim)   :: argument

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  if (iargc() == 0) then   ! There is no input argument.
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!'; stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  else
    do i = 1, iargc()
      call getarg(i, argument)
      if (argument(1:len(keyword)) == keyword) then
        read(argument(len(keyword)+2:len_trim(argument)),*) ivalue
        return
      endif
    enddo
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!';  stop
    endif
    if (present(default_val)) ivalue = default_val   ! Use the default_val value.
  endif
end subroutine
!--------------------------------------------------------------------
subroutine readCmdReal(keyword, fvalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  real,         intent(in), optional :: default_val
  character(*), intent(in)           :: keyword
  real,         intent(out)          :: fvalue
  integer           :: iargc, i
  logical           :: ismissing
  character(cdim)   :: argument

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  if (iargc() == 0) then   ! There is no input argument.
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!';  stop
    endif
    if (present(default_val)) fvalue = default_val   ! Use the default_val value.
  else
    do i = 1, iargc()
      call getarg(i, argument)
      if (argument(1:len(keyword)) == keyword) then
        read(argument(len(keyword)+2:len_trim(argument)),*) fvalue
        return
      endif
    enddo
    if (ismissing)  then
      write(*,*) 'Parameter ', trim(keyword), ' is missing!'; stop
    endif
    if (present(default_val)) fvalue = default_val   ! Use the default_val value.
  endif
end subroutine

!********************************************************
subroutine readParFileString(filename, keyword, cvalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  character(*), intent(in), optional :: default_val
  character(*), intent(in)           :: filename, keyword
  character(*), intent(out)          :: cvalue
  integer          :: i
  character(200)   :: line
  logical          :: ismissing

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  open(10, file=filename, form='formatted', status='old')
  do i = 1,nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      cvalue = line(len(keyword)+2:len_trim(line))
      close(10); return
    endif
  enddo
  100 continue

  if (ismissing) then
    write(*,*) 'Parameter ', trim(keyword), ' is missing!'; close(10); stop
  endif
  if (present(default_val)) cvalue = default_val  ! Use the default_val value.
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileLogical(filename, keyword, lvalue, default_val)
  logical,      intent(in)           :: default_val  !!! must be provided.  Thus no missing.
  character(*), intent(in)           :: filename, keyword
  logical,      intent(out)          :: lvalue
  integer          :: i
  character(200)   :: line

  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      read(line(len(keyword)+2: len_trim(line)),*) lvalue
      close(10); return
    endif
  enddo
  100 continue
  lvalue = default_val  ! Use the default_val value.
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileLogical1(filename, keyword, lvalue, default_val)
  logical,      intent(in)            :: default_val  !!! must be provided.  Thus no missing.
  character(*), intent(in)            :: filename, keyword
  logical(1),   intent(out)           :: lvalue
  integer          :: i
  character(200)   :: line

  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      read(line(len(keyword)+2: len_trim(line)),*) lvalue
      close(10); return
    endif
  enddo
  100 continue
  lvalue = default_val  ! Use the default_val value.
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileInteger1(filename, keyword, ivalue, default_val, is_required)
  logical,      intent(in), optional  :: is_required
  integer,      intent(in), optional  :: default_val
  character(*), intent(in)            :: filename, keyword
  integer(1),   intent(out)           :: ivalue
  integer          :: i
  character(200)   :: line
  logical          :: ismissing

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      read(line(len(keyword)+2:len_trim(line)),*) ivalue
      close(10); return
    endif
  enddo
  100 continue
  if (ismissing) then
    write(*,*) 'Parameter ', trim(keyword), ' is missing!'; close(10); stop
  endif
  if (present(default_val))  ivalue = default_val  ! Use the default_val value.
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileInteger(filename, keyword, ivalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  integer,      intent(in), optional :: default_val
  character(*), intent(in)           :: filename, keyword
  integer,      intent(out)          :: ivalue
  integer          :: i
  character(200)   :: line
  logical          :: ismissing

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      read(line(len(keyword)+2:len_trim(line)),*) ivalue
      close(10); return
    endif
  enddo
  100 continue
  if (ismissing) then
    write(*,*) 'Parameter ', trim(keyword), ' is missing!'; close(10); stop
  endif
  if (present(default_val))  ivalue = default_val  ! Use the default_val value.
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileInteVec(filename, keyword, ivec, nelem, default_vec, is_required)
! if not in xx=[a,b,c] format, treat as 1-element vector
  logical,      intent(in), optional :: is_required
  integer,      intent(in), optional :: default_vec(:)
  integer,      intent(out),optional :: nelem ! # components actually read
  character(*), intent(in)           :: filename, keyword
  integer,      intent(out)          :: ivec(:)
  integer          :: i, head, tail, len_lc
  character(200)   :: line
  logical          :: ismissing

  if (present(nelem))  nelem = 0
  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_vec)
  endif
  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      head = scan(line,'[')
      tail = scan(line,']')
      if (head==0 .or. tail==0) then
        ! write (*,*) 'Vector elements of ', trim(keyword), ' must be enclosed in [ ]'; close(10); stop
        read(line(len(keyword)+2:len_trim(line)),*) ivec(1)
        if (present(nelem)) nelem = 1
        close(10); return
      endif
      call Str2Vector_VaryingLen(line(head+1 : tail-1), ivec, len_lc)
      if (present(nelem)) nelem = len_lc
      close (10); return
    endif
  enddo
  100 continue

  if (ismissing) then
    write(*,*) 'Vector ', trim(keyword), ' is missing!'; close(10); stop
  endif
  if (present(default_vec)) then
    ivec = default_vec  ! Use the default_val values.
    if (present(nelem)) nelem = size(default_vec)
  endif
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileRealVec(filename, keyword, rvec, nelem, default_vec, is_required)
! if not in xx=[a,b,c] format, treat as 1-element vector
  logical,      intent(in), optional :: is_required
  real   ,      intent(in), optional :: default_vec(:)
  integer,      intent(out),optional :: nelem ! # components actually read  
  character(*), intent(in)           :: filename, keyword
  real,         intent(out)          :: rvec(:)
  integer          :: i, head, tail, len_lc
  character(200)   :: line
  logical          :: ismissing

  if (present(nelem))  nelem = 0
  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_vec)
  endif
  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      head = scan(line,'[')
      tail = scan(line,']')
      if (head==0 .or. tail==0) then
        ! write (*,*) 'Vector elements of ', trim(keyword), ' must be enclosed in [ ]'; close(10); stop
        read(line(len(keyword)+2:len_trim(line)),*) rvec(1)
        if (present(nelem)) nelem = 1
        close(10); return
      endif
      call Str2Vector_VaryingLen(line(head+1 : tail-1), rvec, len_lc)
      if (present(nelem)) nelem = len_lc
      close(10);  return
    endif
  enddo
  100 continue

  if (ismissing) then
    write(*,*) 'Vector ', trim(keyword), ' is missing!';  close(10);  stop
  endif
  if (present(default_vec)) then
    rvec = default_vec  ! Use the default_val values.
    if (present(nelem)) nelem = size(default_vec)
  endif  
  close(10)
end subroutine
!------------------------------------------------------------------------------
subroutine readParFileReal(filename, keyword, fvalue, default_val, is_required)
  logical,      intent(in), optional :: is_required
  real,         intent(in), optional :: default_val
  character(*), intent(in)           :: filename, keyword
  real,         intent(out)          :: fvalue
  integer           :: i
  character(200)    :: line
  logical           :: ismissing

  if (.not. present(is_required)) then
    ismissing = .false.
  else
    ismissing = is_required .and. .not. present(default_val)
  endif
  
  open(10, file=filename, form='formatted', status='old')
  do i =1, nline
    read(10, '(a)', end=100) line
    if (line(1:len(keyword)) == keyword) then
      read(line(len(keyword)+2:len_trim(line)),*) fvalue
      close(10);  return
    endif
  enddo
  100 continue

  if (ismissing) then
    write(*,*) 'Vector ',   trim(keyword), ' is missing!';   close(10);  stop
  endif
  if (present(default_val)) fvalue = default_val  ! Use the default_val value:
  close(10)
end subroutine
!--------------------------------------------------------
END MODULE

