!#define DEBUGr

module myRand_module
! Sample usage:
! call init_random_seed(in:: id)
! call randn(out:: s/v/m)
! rndIntgr(1:N) = randperm(N)
! rndIntgr(1:ni)= rand_unif_spread(N, ni)
! call indexx(in:: rndArray, out:: index_vec)
  USE nrtype
  USE nrutil, ONLY : array_copy, arth, assert_eq, nrerror, swap
  USE myMath_module
  IMPLICIT  NONE
  
  INTERFACE indexx
    module procedure indexx_sp
    module procedure indexx_i4b
  END INTERFACE
  
  INTERFACE randb ! randb(hrvst), integer(:) Binomial(0,1)
    module procedure randb_v, randb_m
  END INTERFACE
  
  INTERFACE randn ! N(0,1)
    module procedure randn_s
    module procedure randn_v
    module procedure randn_m
  END INTERFACE 
  
  INTERFACE randperm 
    module procedure randperm_full, randperm_subset
  END INTERFACE 
  
  INTERFACE recurPartiPerm 
    module procedure recurPartiPerm_full, recurPartiPerm_subset
  END INTERFACE

  INTEGER, PARAMETER :: NquasiRand= 5003
  REAL   , DIMENSION(NquasiRand) :: rand_qs_v 
  LOGICAL(1) :: rands_read_already = .false.
! Matlab generate a prime number of quasi-random #'s:
!  p = haltonset(1,'Skip',10007,'Leap',700)  % 1D.  10007 & 701 are prime
!  p = scramble(p,'RR2');
!  N = 5003;     % prime #
!  x = net(p,N); Tab = randperm(N); x = x(Tab);
 PRIVATE   :: NquasiRand, rand_qs_v, rands_read_already
CONTAINS
!==============================
subroutine read_quasi_random_numbers
 open(10, file='rand_5003.bin', access='direct', err=20, recl=I4*NquasiRand)
 read(10, rec=1) rand_qs_v
 close(10)
 return
20 stop ' Read error of rand_5003.bin!'
end subroutine
!-------------
subroutine rand_quasi(x, NPi, RANKi)
real, intent(out):: x(:)
INTEGER, OPTIONAL, INTENT(IN) :: NPi, RANKi  ! # proc in MPI, rank of the calling proc.
! in MPI case, each proc only reads out (1+rank + k*np)'th elements, k=0,1,2,...
! 'coz we certainly don't want every proc reads the same set of random numbers.
integer, save    :: ptr = 1   ! formally init.,  ptr points to the next slot
integer  ::  nx, k, np, rank

np=1; rank=0
if (present(NPi)) then
  np = NPi;  rank = RANKi
end if
if (.not. rands_read_already) then  ! 1st time calling
  call read_quasi_random_numbers
  ptr = rank + 1  ! as rank=0,1,2,..  real init.
  rands_read_already = .true.
endif
nx = size(x)
do k = 1, nx
  x(k)= rand_qs_v(ptr)
  ptr = ptr + np
  if (ptr>NquasiRand) ptr = ptr - NquasiRand
end do
end subroutine
!-------------
subroutine reverse_gener_Mapping(ind, rvs, rvsn)
! ind   N x 1  int, ind: I=[1,...,N] ---> V={1,...,P} is a mapping
! rvs   P x m  int, rvs: V ---> I  is the inverse mapping of ind,
!                   rvs(k,:) = Ik = {all points \in I that is mapped to k}
!       make sure m is big enough to hold Ik, \forall k \in V
! rvsn  P x 1  int, rvsn(k)  = Cardinality of Ik (must <= m)
integer, intent(in) :: ind(:)
integer, intent(out):: rvs(:,:), rvsn(:)
integer  :: N, P, M, i, j
N = size(ind)
P = size(rvs, 1);  M = size(rvs, 2)
if (P /= size(rvsn)) stop 'rvsn size must = # rows in rvs. In reverse_gener_Mapping()'
rvsn = 0;          rvs = 0   ! init.
do i = 1,N
  j  = ind(i)
  if (j > P)       stop 'ind value exceeds # rows in rvs. In reverse_gener_Mapping()'
  rvsn(j) = rvsn(j) + 1
  if (rvsn(j) > M) stop '# elem. exceeds # cols in rvs. In reverse_gener_Mapping()'
  rvs(j, rvsn(j)) = i
enddo
end subroutine

subroutine reverse_11_Mapping(ind, rvs)
! ind  N x 1  int, ind: I={1,...,N} ---> I, is a 1-to-1 mapping.
!            E.g., ind = [4 1 3 2]  (i.e., [1 2 3 4] --> [4 1 3 2])
! rvs  N x 1  int, reverse mapping of ind
!            E.g., rvs = [2 4 3 1]  (i.e., [1 2 3 4] --> [2 4 3 1])
! Since, e.g., ind(4)=2, we have rvc(2)=4. That is,
! ind(I) = ind.  So, rvs(ind) = I.
integer, intent(in) :: ind(:)
integer, intent(out):: rvs(:)
integer  :: N, k
N = size(ind)
rvs(ind) = (/ (k, k = 1,N) /)
end subroutine

subroutine reverse_10_Mapping(ind, rvs, shf)
! ind  N x 1  int, ind: I={1,...,N} ---> E={0,1,..,M}, where M < N.
! E.g., ind = [0 4 1 0 0 3 2 0].  
! Here, N = 8 (cardinality of I, or, length of ind);  M = 4.
! N-M elements of I are mapped to 0, which denotes invalid number.
! The nonzero mappings are nonetheless 1-to-1.
! rvs  M x 1  int, reverse mapping of the nonzero subset of ind.
!      (We need to know M beforehand.)
! Since, e.g., ind([2 3 6 7]) = [4 1 3 2], rvs([4 1 3 2]) = [2 3 6 7],
! i.e., rvc() = [3 7 6 2].
! New addition:
! ind: I --> E1 = {0, shf+{1,...,M}}
! rvs: still {1,...,M} --> {1,...,M}, knowing implicitly that it's
!        shf+{1,...,M} --> {1,...,M}.
integer, intent(in) :: ind(:)
integer, intent(in), optional :: shf
integer, intent(out):: rvs(:)
integer  :: subi(size(rvs)), subo(size(rvs))  ! e.g., subi=[2 3 6 7], subo=[4 1 3 2]
integer  :: N, M, k, i
N = size(ind)
M = size(rvs)
i = 0
do k = 1,N
  if (ind(k)==0) cycle  ! skip
  i = i+1
  if (i > M) stop '# nonzero elements > size(rvs). In reverse_10_Mapping().'
  subi(i) = k
  subo(i) = ind(k)
enddo
if (present(shf))  subo = subo - shf
if (i < M) stop '# nonzero elements < size(rvs). In reverse_10_Mapping().'
rvs(subo) = subi
end subroutine
!-------------------------------
SUBROUTINE init_random_seed(ID)
  INTEGER, INTENT(IN), OPTIONAL :: ID  ! such as a parallel
  ! process' ID, so that different process has different
  ! pseudo-random sequence.
  INTEGER  :: n, clock, id00
  INTEGER, DIMENSION(40), PARAMETER:: prime = (/&
   7573,  7577,  7583,  7589,  7591,  7603,  7607,  7621,  7639,  7643,&
   7649,  7669,  7673,  7681,  7687,  7691,  7699,  7703,  7717,  7723,&
   7727,  7741,  7753,  7757,  7759,  7789,  7793,  7817,  7823,  7829,&
   7841,  7853,  7867,  7873,  7877,  7879,  7883,  7901,  7907,  7919/) 

  CALL RANDOM_SEED
  CALL RANDOM_SEED(SIZE  = n)
  CALL SYSTEM_CLOCK(COUNT= clock)
#ifdef DEBUGd
  write(*,*) 'clock count:', clock; call flush(6)
#endif
  IF (PRESENT(ID)) THEN
    id00 = ID
  ELSE
    id00 = 0
  ENDIF
  CALL RANDOM_SEED(PUT = clock - 100019*id00 + 10007*prime(1:n) )
END SUBROUTINE init_random_seed

!*************************************
 
  SUBROUTINE randb_v(harvest)  ! binary (0,1)
  integer, INTENT(OUT) :: harvest(:)
  real     :: tmp(1024)   ! this is to reduce the size of tmp vector
  integer  :: m, rem, i,  head
  rem = size(harvest)
  m   = rem / 1024;   rem = rem - m * 1024
  do i = 1, m
    call RANDOM_NUMBER(tmp) 
    head = (i-1)*1024 + 1
    harvest(head : head + 1023) = floor(2*tmp)
  enddo
  if (rem /= 0) then
    call RANDOM_NUMBER(tmp(1:rem))
    head = m*1024 + 1
    harvest(head : head + rem-1) = floor(2*tmp(1:rem)) 
  endif
  END SUBROUTINE
  
  SUBROUTINE randb_m(harvest)  ! binary (0,1)
  integer, INTENT(OUT) :: harvest(:,:)
  real     :: tmp(size(harvest,1),size(harvest,2))

  call RANDOM_NUMBER(tmp)  ! quik and easy, w/o memory consideration
  harvest  = floor(2*tmp) 
  END SUBROUTINE
  
  SUBROUTINE randn_s(harvest)  ! scalar N(0,1)
  ! modified from NumRecipe's gas_dev..
  ! Here I use f90 intrinsic subroutine RANDOM_NUMBER() 
  ! instead of NR's rand1.
  REAL(SP), INTENT(OUT) :: harvest
  REAL(SP)  :: rsq,v1,v2
  REAL(SP), SAVE :: g
  LOGICAL,  SAVE :: gaus_stored=.false.
  if (gaus_stored) then
    harvest = g
    gaus_stored = .false.
  else
    do
      call RANDOM_NUMBER(v1)  ! use f90 intrinsic func
      call RANDOM_NUMBER(v2)
      v1 = 2.0_sp*v1 - 1.0_sp
      v2 = 2.0_sp*v2 - 1.0_sp
      rsq= v1**2 + v2**2
      if (rsq > 0.0 .and. rsq < 1.0) exit
    end do
    rsq  =   sqrt(-2.0_sp * log(rsq)/rsq)
    harvest= v1*rsq
    g    =   v2*rsq
    gaus_stored= .true.
  end if
  END SUBROUTINE randn_s
  
! ---------------------------------
  SUBROUTINE randn_v(harvest)
  REAL(SP), DIMENSION(:), INTENT(OUT) :: harvest
  REAL(SP), DIMENSION(size(harvest)) :: rsq,v1,v2
  REAL(SP), ALLOCATABLE, SAVE :: g(:)
  INTEGER(I4B) :: n,ng,nn,m
  INTEGER(I4B), SAVE :: last_allocated = 0
  LOGICAL,  SAVE :: gaus_stored = .false.
  LOGICAL   :: mask(size(harvest))
  n = size(harvest)
  if (n /= last_allocated) then
    if (last_allocated /= 0) deallocate(g)
    allocate(g(n))
    last_allocated = n
    gaus_stored=.false.
  end if
  if (gaus_stored) then
    harvest    = g
    gaus_stored=.false.
  else
    ng = 1
    do
      if (ng > n) exit
      call RANDOM_NUMBER(v1(ng:n))
      call RANDOM_NUMBER(v2(ng:n))
      v1(ng:n)  = 2.0_sp*v1(ng:n)-1.0_sp
      v2(ng:n)  = 2.0_sp*v2(ng:n)-1.0_sp
      rsq(ng:n) = v1(ng:n)**2+v2(ng:n)**2
      mask(ng:n)= (rsq(ng:n)>0.0 .and. rsq(ng:n)<1.0)
      call array_copy(pack(v1(ng:n),mask(ng:n)),v1(ng:),nn,m)
      v2(ng:ng+nn-1) =pack(v2(ng:n),mask(ng:n))
      rsq(ng:ng+nn-1)=pack(rsq(ng:n),mask(ng:n))
      ng= ng+nn
    end do
    rsq = sqrt(-2.0_sp * log(rsq)/rsq)
    harvest = v1*rsq
    g   = v2* rsq
    gaus_stored = .true.
  end if
  END SUBROUTINE randn_v
  
! --------------------------------- 
  SUBROUTINE randn_m(harvest) ! for harvest being a 2d matrix
  REAL(SP), INTENT(OUT) :: harvest(:,:)
  REAL(SP), ALLOCATABLE :: temp(:)   ! 1d vector
  INTEGER(I4B) :: sz(2)  ! size of harvest
  sz  = shape(harvest)
  allocate(temp(sz(1)*sz(2)))
  call randn_v(temp)  ! within one file, randn_v is still visible
  harvest = reshape(temp, sz)
  deallocate(temp)
  END SUBROUTINE randn_m

!==============================
function randperm_full(N)
  integer(I4B),  intent(in) :: N
  integer(I4B):: randperm_full(N)
  REAL(SP)    :: arr(N)
  call random_number(arr)
  call indexx(arr, randperm_full) ! sorting
end function randperm_full

function randperm_full_accumu(N)  result(randperm_full)
! accumulative randperm, performing on top of previous permutations
  integer(I4B),  intent(in) :: N
  integer(I4B):: randperm_full(N)
  REAL(SP)    :: arr(N)
  logical, save::firstEntry = .true.
  integer(I4B), save        :: cacheSize
  integer(I4B), save, allocatable :: cache(:) ! remember last time's randperm
  integer(I4B)  :: k
  
  call random_number(arr)
  call indexx(arr, randperm_full) ! sorting
  if (firstEntry) then
    firstEntry = .false.
    cacheSize  = N
    allocate(cache(N))
    cache = (/ (N+1-k, k = 1,N) /)
#ifdef DEBUGd
   write(*,*) 'cache', cache;  call flush(6)
#endif
  else if (cacheSize /= N) then
    deallocate(cache)
    cacheSize  = N
    allocate(cache(N))
    cache = (/ (N+1-k, k = 1,N) /)
#ifdef DEBUGd
   write(*,*) cacheSize, size(cache), N;  call flush(6)
#endif    
  endif
#ifdef DEBUGd
  if (size(cache) /= size(randperm_full)) then
    write(*,*) 'size(cache; randperm_full), N, cacheSize:', size(cache), size(randperm_full), N, cacheSize; call flush(6)
  endif
#endif
  randperm_full = randperm_full(cache)
  cache   =  randperm_full
end function randperm_full_accumu

function randperm_subset(N, M)
! shuffle 1,...,N, yet output the 1st M numbers only (M <= N)
  integer(I4B), intent(in)  :: N, M
  integer(I4B)  :: randperm_subset(M), arr(N) 
  arr = randperm_full(N)
  randperm_subset = arr(1:M)   ! pick the 1st M entries
end function randperm_subset

!==============================
function partiPerm(q, m)
! Permute q to form p (q & p are vectors of length N):
!   (m is a factor of N.)
! p is divided into m partitions.  In each of N/m turns, grab m
!   consecutive entries out of q and assign them in permutation
!   to the m partitions of p.  
! Do this from low end to high of q; q is sorted in ascending order,
! so, in each partition of p, the entries are also sorted in
!   ascending order.
integer :: q(:), partiPerm(size(q)), m, rnd(m), N, W, k, i, beg
N = size(q)
W = N/m
if (W*m/=N) then
 write(*,*) m, 'must be a factor of ', N;  call flush(6);  stop '!'
endif
do k = 1, W     ! turns
  rnd= randperm(m)
  beg= (k-1)*m  ! q's chunk's beginning-1
  do i = 1, m
    partiPerm((i-1)*W + k) = q(beg+rnd(i))
  enddo
enddo
end function
!--
function recurPartiPerm_full(N)  result(permu)
! permute 1,...,N, by hierarchical partition-permutation
! MOTIVATION: make this permutation of low discrepancy, i.e., avoid 
! clustering of nearby values at nearby indices (of the resulting 
! permutation vector).  Spread values out, evenly.
! Note:
! Although in developing this code what I had in mind was:
!  freq. is along y-axis;  source is along x-axis;
!  assign freq. to source (or, equiv. map source to freq.), while in 
!  actual usage it's the other way around, fortunately,
! both the forward and reverse mapping of recurPartiPerm() should have 
! similar performance of low discrepancy.  
! Therefore, this can equally be regarded as a permutation for mapping 
! freq. to source.
 
integer :: N, permu(N), fac(ceiling(log(real(N))/log(2.0))), Nfac, &
           Nparti, W, k, j, beg
           
 permu = (/ (k, k = 1,N) /)! init.
 call PRIMEFACTORS(N, fac, Nfac)           
 Nparti = 1;    W = N
 do k = 1, Nfac     ! level k
   do j = 1, Nparti ! partitions on level k
     beg= (j-1)*W
     permu(beg+1 : beg+W) = partiPerm(permu(beg+1 : beg+W), fac(k));
   enddo
   Nparti = Nparti * fac(k)
   W      = W / fac(k)
 enddo  
end function
!--
function recurPartiPerm_subset(N, nsub)  result(permu)
! permute 1,...,N, by hierarchical partition-permutation
! then only pick nsub entries out of the permuted vector
integer :: N, nsub, permu(nsub), tmp(N)
 tmp   = recurPartiPerm_full(N)
 permu = tmp(1:nsub)
end function

!---------------------------
! This block of functions use repelling potential to spread out 
! frequency-to-source mapping, i.e., nearby frequency values shouldn't be 
! mapped to nearby source locations.
! The potential is defined as 1/r_ij, where r_ij is the 'distance' between i & j.
! This distance is Euclidean distance in 2D: |(si-sj, fi-fj)|, where fi & fj are
! frequences (like y-axis), and si & sj are sources (like x-axis).
! [To avoid boundary effect, use periodic boundary condition in computing (si-sj).
!  Otherwise, low and high ends of source indices tend to have medium frequency,
!  because this would make overall potential energy smaller.]
! Note:
! Here, the x-axis is source while the y-axis is freq.  So it's mapping source 
! to freq.  But in final usage we want mapping freq. to source.  For this we 
! just need to invoke reverse mapping.
!
! Once the potential, or energy, is defined, we use greedy algorithm to find
! a very good configuration starting from randperm().
! [Note that starting from a good one may make this algorithm slower, because
!  a lot of tries for improvement would fail.  That's why we just start from
!  randperm() instead of recurPartiPerm().]
! 
! To handle the case where only a subset of permuted numbers are selected, 
! '0' is used as dummy padding to denote empty entries (to keep the whole vector
! length the same.)
! 
! Multiple permutations (on different segments of numbers) can coexist.
! The current permutation's energy is affected by existing ones.  We can
! incorporate this fact easily by taking account of the interaction energy. 
! The multiple permutations are repesented in a matrix N X m,
! (for m runs of N-permutation.)
! 2011/08/15
function  poten_ij(i,j,pi,pj, N,m)
! pi, pj  1 x m vectors, each containing m freq values.
! the respective last (i.e., mth) component is the one in interest.
! return Coulumb potential, felt at pi(m), due to pj
! 1-->m-1 components are what have been selected earlier, conditioned
! on which we ajust the latest component to minimize the energy.
! whenever a p(k)==0, meaning this component is empty, to be ignored.
integer ::  i, j, m, N, pi(:), pj(:), k, dij2, dij2b, jb
real    ::  poten_ij, ra2, rb2
if (pi(m)==0) then  ! empty dummy entry
   poten_ij = 0.
   return
endif
if (j>i) then    ! periodic extension
   jb = j - N
else        
   jb = j + N
endif
dij2  = (i - j)**2
dij2b = (i - jb)**2
if (i/=j .and. pj(m)/=0) then
  ra2   = (pi(m) - pj(m))**2 + dij2
  rb2   = (pi(m) - pj(m))**2 + dij2b
  poten_ij = 1./sqrt(ra2) + 1./sqrt(rb2)
else
  poten_ij = 0   ! no self-energy
endif
do k = 1, m-1
  if(pj(k)==0)  cycle
  ra2= (pi(m) - pj(k))**2 + dij2
  rb2= (pi(m) - pj(k))**2 + dij2b
  poten_ij = poten_ij + 1./(0.01+sqrt(ra2)) + 1./(0.01+sqrt(rb2)) ! 0.01 because
  ! it may happen that i==j & pi(m)==pj(k), then r==0.  to avoid 1/0.
enddo
end function

function   poten_i(i,p,N,m)  result(Ei)
real    :: Ei    ! potential energy felt at p(i,m)
integer :: i, p(:,:), N, m, k, bg, ed, bg2, ed2
integer, parameter :: W2 = 16

Ei = 0.
if (p(i,m)==0)return

bg2 = 1;      ed2= 0
if (N < 50) then
  bg = 1;     ed = N;
else
  bg = i-W2;  ed = i+W2  ! limited domain
  if (bg<=0) then
    bg2= bg+N;ed2= N
    bg = 1
  else if (ed>N) then
    bg2= 1;   ed2= ed-N
    ed = N
  endif
endif

do k = bg,ed
  Ei = Ei + poten_ij(i,k,p(i,:),p(k,:),N,m)
enddo
do k = bg2,ed2
  Ei = Ei + poten_ij(i,k,p(i,:),p(k,:),N,m)
enddo
end function

function try_swap_ij(Ei,Ej,i,j,p, N,m) result(suc)
! swap p(i,m) & p(j,m), dE?
! if dE <0, then this move is accepted
! Ei, Ej:   potential energy associated w/ p(i,m), p(j,m) 
 logical suc
 real,   intent(inout) :: Ei, Ej
 integer,intent(inout) :: p(:,:)
 integer,intent(in)    :: i, j, N, m
 real    :: Oi, Oj
 integer :: itmp
 itmp=p(i,m);  p(i,m)=p(j,m); p(j,m)=itmp ! trial swap
 Oi  = poten_i(i,p,N,m)
 Oj  = poten_i(j,p,N,m)
 if (Oi+Oj < Ei+Ej) then
   Ei = Oi;  Ej = Oj    ! accept
   suc= .true.
 else ! reject the move, swap back pi & pj:
   itmp=p(i,m);  p(i,m)=p(j,m); p(j,m)=itmp
   suc= .false.
 end if

end function

subroutine pair_optim_poten(p, N,m)
! attempt swapping pairs of freq's, trying to reduce the potential energy
integer, intent(in)    :: N, m
integer, intent(inout) :: p(:,:)  ! N x m
integer  :: i,j, nfail, tmp
real     :: Y, E(N)
#ifdef DEBUGr
 integer :: SUCC
 SUCC = 0
 write(*,*) 'm=', m, ' before optim: p(:,m)='
 write(*,*)  p(:,m)
#endif

do i = 1, N
  E(i)= poten_i(i,p,N,m)  ! init. energy
end do
nfail = 0
do while (nfail < N)
  CALL RANDOM_NUMBER(Y); Y=Y*0.999+5e-4;  i = ceiling(N*Y)
  j = i
  do while(j==i)
    CALL RANDOM_NUMBER(Y); Y=Y*0.999+5e-4;  j = ceiling((N-1)*Y)
  end do
  if (try_swap_ij(E(i),E(j),i,j,p, N,m)) then
    nfail = 0
#ifdef DEBUGr
    SUCC = SUCC + 1
#endif    
  else
    nfail = nfail + 1
  endif
end do
#ifdef DEBUGr
 write(*,*) 'm, #success=', m, SUCC, ' After optim:'
 do i = 1,m
   write(*,*) p(:,i)
 enddo
 call flush(6)
#endif
end subroutine

function rand_unif_spread2(Nsrc, nfrq)  result(assSrc)
! see rand_unif_spread() for rationale.
! This approach: setup an energy, to reduce this energy is equiv. to
! spreading out the . assignment.
integer(I4B), intent(in)    :: Nsrc, nfrq
integer(I4B)  :: assSrc(nfrq), Nturns, rem, k, done
integer(I4B), allocatable   :: p(:,:)
Nturns = int(nfrq / Nsrc)
rem    = nfrq - Nturns*Nsrc
k      = Nturns
if (rem>0)   k = k+1
allocate( p(Nsrc, k) )
do k = 1, Nturns
  done = (k-1)*Nsrc
  p(:,k) = randperm(Nsrc) + done
  call pair_optim_poten(p, Nsrc, k)
  call reverse_11_Mapping(p(:,k)-done, assSrc(done+1: done+Nsrc))  !!!
enddo
if (rem<=0) then
  deallocate(p)
  return
endif  ! do the remainder:
done   = Nturns * Nsrc
k      = Nturns + 1
p(:,k) = randperm(Nsrc)
where (p(:,k) > rem)
  p(:,k) = 0
elsewhere
  p(:,k) = p(:,k) + done  ! since freq. has higher values
endwhere
call pair_optim_poten(p, Nsrc, k)
call reverse_10_Mapping(p(:,k), assSrc(done+1: ), done)
deallocate(p)
end function
!---------------------------
function rand_unif_spread(Nsrc, nfrq, LOWDISCREP)  result(gotSrcIdx)
! gotSrcIdx     nfrq x 1,    each 'freq' component gets a 'source'
! index, out of Q ={1,..,Nsrc}.  This is also a mapping, from a set 
! F = {1,...,nfrq} (like x-axis) to Q (like y-axis).
! This mapping (or, assignment) aims to be uniform.  That is, don't map
! nearby freq's to nearby sources.  Or, don't assign nearby sources to
! nearby freq's.
! Typically, Nsrc < nfrq, then, make several runs.  In each run, pick a 
! segment of length Nsrc in F, and assign Nsrc source indices in
! permutation to those Nsrc contiguous elements in F. 
! (A source may end up getting several freqs, if Nsrc < nfrq.  But a 
! freq is mapped to one and only one source.)
!
! LOWDISCREP==1  try to achieve an assignment.mapping that is of low
!   discrepancy, i.e., to really evenly spread out the mapping.
!   This requires a technique more involved than the naive randperm().
! The technique I used is called 'hierarchical partition-permutation'.
! LOWDISCREP==2  use repelling potential, rand_unif_spread2()
! 
! See  recurPartiPerm(), and randDraw_wo_replace.m

  integer(I4B), intent(in)       :: Nsrc, nfrq
  integer, intent(in), optional  :: LOWDISCREP
  integer(I4B), dimension(nfrq)  :: gotSrcIdx
  integer(I4B)  :: Nturns, rem, k, done
  logical       :: EVEN
  EVEN = .false.   ! default naive approach
  if (present(LOWDISCREP))  then
    if (LOWDISCREP .eq. 2)  then
      gotSrcIdx = rand_unif_spread2(Nsrc, nfrq)
      return
    endif
    EVEN = LOWDISCREP .eq. 1
  endif
  Nturns = int(nfrq / Nsrc)
  rem    = nfrq - Nturns*Nsrc
  do k   = 1, Nturns
    done = (k-1)*Nsrc
    if (EVEN) then
      gotSrcIdx(done+1 : done+Nsrc) = recurPartiPerm(Nsrc)
    else
      gotSrcIdx(done+1 : done+Nsrc) = randperm(Nsrc)
    endif
  end do
  if (rem > 0)  then
    done = Nturns * Nsrc
    if (EVEN) then
      gotSrcIdx(done+1 : ) = recurPartiPerm(Nsrc, rem) 
    else
      gotSrcIdx(done+1 : ) = randperm(Nsrc, rem)  ! not biased toward small #'s
    endif
  end if
end function rand_unif_spread
!-----
subroutine rand_unif_condense(N, drawer, nelem)
! UNIFORMLY condense D={1,..,N} into P drawers. P must <= N.
! Each number is assigned to one and only one drawer.
! Each drawer's # of elements <= ceil(N/P)
! N       1 x 1
! drawer  P x M
! nelem   P x 1
integer, intent(in) :: N
integer, intent(out):: drawer(:,:), nelem(:)
integer  :: P, ind(N)
P   = size(drawer, 1)
ind = rand_unif_spread(P, N)
call  reverse_gener_Mapping(ind, drawer, nelem)
end   subroutine

! *********************************
  SUBROUTINE indexx_sp(arr,index)
  REAL(SP), DIMENSION(:), INTENT(IN) :: arr
  INTEGER(I4B), DIMENSION(:), INTENT(OUT) :: index
  INTEGER(I4B), PARAMETER :: NN=15, NSTACK=50
  REAL(SP)     :: a
  INTEGER(I4B) :: n,k,i,j,indext,jstack,l,r
  INTEGER(I4B), DIMENSION(NSTACK) :: istack
  n=assert_eq(size(index),size(arr),'indexx_sp')
  index=arth(1,1,n)
  jstack=0
  l=1
  r=n
  do
    if (r-l < NN) then
      do j = l+1, r
        indext=index(j)
        a  = arr(indext)
        do i=j-1,l,-1
          if (arr(index(i)) <= a) exit
          index(i+1)=index(i)
        end do
        index(i+1)=indext
      end do
      if (jstack == 0) RETURN
      r= istack(jstack)
      l= istack(jstack-1)
      jstack= jstack-2
    else
      k= (l+r)/2
      call swap(index(k),index(l+1))
      call icomp_xchg(index(l),index(r))
      call icomp_xchg(index(l+1),index(r))
      call icomp_xchg(index(l),index(l+1))
      i= l+1
      j= r
      indext= index(l+1)
      a= arr(indext)
      do
        do
          i = i+1
          if (arr(index(i)) >= a) exit
        end do
        do
          j = j-1
          if (arr(index(j)) <= a) exit
        end do
        if (j < i) exit
        call swap(index(i),index(j))
      end do
      index(l+1)= index(j)
      index(j)  = indext
      jstack = jstack+2
      if (jstack > NSTACK) call nrerror('indexx: NSTACK too small')
      if (r-i+1 >= j-l) then
        istack(jstack)  = r
        istack(jstack-1)= i
        r= j-1
      else
        istack(jstack)  = j-1
        istack(jstack-1)= l
        l= i
      end if
    end if
  end do
  CONTAINS
!BL
    SUBROUTINE icomp_xchg(i,j)
      INTEGER(I4B), INTENT(INOUT) :: i,j
      INTEGER(I4B) :: swp
      if (arr(j) < arr(i)) then
        swp= i
        i  = j
        j  = swp
      end if
    END SUBROUTINE icomp_xchg
  END SUBROUTINE indexx_sp
  
! ---------------------------------

  SUBROUTINE indexx_i4b(iarr,index)
  INTEGER(I4B), DIMENSION(:), INTENT(IN) :: iarr
  INTEGER(I4B), DIMENSION(:), INTENT(OUT) :: index
  INTEGER(I4B), PARAMETER ::  NN=15, NSTACK=50
  INTEGER(I4B) :: a
  INTEGER(I4B) :: n,k,i,j,indext,jstack,l,r
  INTEGER(I4B), DIMENSION(NSTACK) :: istack
  n = assert_eq(size(index),size(iarr),'indexx_sp')
  index = arth(1,1,n)
  jstack= 0
  l = 1
  r = n
  do
    if (r-l < NN) then
      do j  = l+1, r
        indext= index(j)
        a   = iarr(indext)
        do i= j-1, l, -1
          if (iarr(index(i)) <= a) exit
          index(i+1) = index(i)
        end do
        index(i+1)= indext
      end do
      if (jstack == 0) RETURN
      r= istack(jstack)
      l= istack(jstack-1)
      jstack= jstack-2
    else
      k= (l+r)/2
      call swap(index(k),index(l+1))
      call icomp_xchg(index(l),index(r))
      call icomp_xchg(index(l+1),index(r))
      call icomp_xchg(index(l),index(l+1))
      i= l+1
      j= r
      indext= index(l+1)
      a= iarr(indext)
      do
        do
          i = i+1
          if (iarr(index(i)) >= a) exit
        end do
        do
          j = j-1
          if (iarr(index(j)) <= a) exit
        end do
        if (j < i) exit
        call swap(index(i),index(j))
      end do
      index(l+1)= index(j)
      index(j)  = indext
      jstack    = jstack+2
      if (jstack > NSTACK) call nrerror('indexx: NSTACK too small')
      if (r-i+1 >= j-l) then
        istack(jstack)  = r
        istack(jstack-1)= i
        r = j-1
      else
        istack(jstack)  = j-1
        istack(jstack-1)= l
        l = i
      end if
    end if
  end do
  CONTAINS
!BL
    SUBROUTINE icomp_xchg(i,j)
      INTEGER(I4B), INTENT(INOUT) :: i,j
      INTEGER(I4B) :: swp
      if (iarr(j) < iarr(i)) then
        swp= i
        i  = j
        j  = swp
      end if
    END SUBROUTINE icomp_xchg
  END SUBROUTINE indexx_i4b
! *************************************

END MODULE myRand_module

