!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     datatype (like C++ classes) definition
!-----------------------


MODULE datatype0
! the implementations are in  datatype1
use nrtype
implicit none
!=========================================
type option_param
  integer(1) :: CGreset,   CGresetBlock,  FREESURF,    FWI,       &
                MUTE,      SINU,          TOW,         PRECND,    &
                FWD_sbsmp, Derr_mode
end type
! CGreset:      every so many iter's, reset to steep descnt
! CGresetBlock: Once so many resets have taken place, do a line search along the
!               direction:  x_cur - x_{before the block}.  This trick aims at
!               averaging out the zigzag path due to stochastic gradients, and is
!               also like the Gauss-Newton method whose direction is - H^{-1} g
! FREESURF:0 -- no free surface atop,  use pml instead
!          1 -- use free surface atop
!          >1 - pad adptive layer atop; let fwi yield the effective free surface.
!               The benefit is: fwi can correct sources' radiaion pattern.
!               The pad width is FREESURF (must < 127)
! FWI:     0 -- just forward modeling
!          1 -- do fwi
! MUTE:    0 -- no muting 
!          1 -- simple muting of direct wave, throwing away anything 
!               sooner than water bottom reflection, requiring waterbottom
!          2 -- run simulations for 2 nt; keep the 2nd nt (for freq. division)
!          3 -- keep first arrival only.
! SINU:    0 -- use time domain wavelet
!          1 -- use freq. selection, feeding sinusoidal waves
! TOW:     0 -- fixed geometry
!          1 -- towed
!          2 -- two-sided towed  (obtained by using reciprocity & interpolation,
!               so that receivers are: -far-offset  -->  far-offset, where 
!               far-offset = slead + saptr - 1 
! PRECND:  0 -- no preconditioning
!          1 -- z**2     
!          2 -- Sum_s |Laplace. G(x|s)|^2
! FWD_sbsmp  -- forward modeling, save the data subsampled by FWD_sbsmp
! Derr_mode 0-- misfit = csg_calc - csg_obs
!           1--        = alph * csg_calc - csg_obs, with optimal alpha
!           2-- my new trick

type domain_param
  integer    :: lx,   npml,  nz,   nx,   nt,                    &
                z1,   z2,    z3,   z4,   z5,   z6,              &
                x1,   x2,    x3,   x4,   x5,   x6
  real       :: dx,   dt,    vmax,       vmin
end type

type source_param
  integer    :: nw,   nbtot, ncb         ! total # of bundles, # clans per bundle
  real       :: w1,   dw,    f0
  integer    :: s1,   ds,    ns,   zs, & ! sources' x coordi:
                g1,   dg,    ng,   zg, & ! s1 + (i-1)*ds,  i = 1,..,ns.  unit: dx
                slead,  saptr,         &
                z_fix_v,               &
                mskWinLen1,  MaxSrcDelay   
  real,         allocatable  ::    waterBtm(:), s(:), sw(:), phw(:), trctmp(:)
end type        
! if z_fix_v <=0, no water velocity fixing
! else,   if present(waterBtm),  [z_fix_v, waterBtm]: fix velo to vmin
! else,   [1, z_fix_v]:  fix velo to vmin

type io_param_typ
  logical           ::V_TRU_AVAIL,& !  true velo. file on disk
                      XFAST         !  fast dim is horizontal
  character(len=100)::vel_ini_file, vel_out_file, vel_tru_file, &
                      vel_Err_file, dat_Err_file, grd_ite_file, &
                      csgin_file,   csgout_file,  trvlTim_file, &
					  csgfile_ext,                              & 
                      waterbottom_file,           srcwvlt_file, &
                      freq_encod_file
end type   ! csgfile_ext: CSG file extension.  CSG file format:
! e.g. csg??ext, where ?? denotes the shot number; 'csg' denotes
! the original csgin_file.  Note, csgout_file will append '.bin' always

type entire_param
  type(option_param) :: o
  type(domain_param) :: d
  type(source_param) :: s
  type(io_param_typ) :: i
end type

!^^^^^^^^^^^^^^^
! 'r':  receiver     'i':  global index
!                    'j':  global index (in the intersected set because of the clan) 
!                          mnemonic: 'jiao-'  is 'intersection' in Chinese
! 's':  source       '1':  the 1st (global index)
! 'c':  clan         '2':  the 2nd (global index)
! 'b':  bundle
type source_type
  integer             :: ri1, ri2   ! associated receivers' 1st & last index, global.
  real, allocatable   :: csgobs(:,:)! observed csg storage, for freq.selec. only
end type
!.............
type clan_type
  type(source_type),  allocatable  ::  s(:)
  integer             :: si1, si2, &
                         ri1, ri2, &! non-intersected
                         rj1, rj2   ! the 1st & last receiver index in the common
end type                            !  (i.e., intersected) aperture of clan
!.............
type bund_type
  integer             :: si1, si2, &
                         ri1, ri2, & ! non-intersected original receiver range                
                         rj1, rj2, & ! receiver index range intersected due to clan
						 nc          ! # clans in this bundle
  type(clan_type),    allocatable  ::  c(:)
  integer,            allocatable  ::  pol(:), dly(:), ci_von_nw(:)  ! for encoding
  real,               allocatable  ::  rndtmp(:)
end type
!.............
type csg_current       ! computed, observed, {error/temp}, 
  real,allocatable::  c(:,:), o(:,:), e(:,:)
end type               ! for the proc at current iteration
! . . . . .

type proc_type         ! data on a process
  integer             ::  bi1, bi2  !  global bundle index
  type(bund_type),    allocatable  ::  b(:)
  type(csg_current)   ::  csg   
end type
!^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
END MODULE
