!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     datatype (like C++ classes) implementation
!-----------------------


MODULE datatype1
! the definitions are in  datatype0
use nrtype;                use myMPI0
use myFFTW0;               use myRand_module
use myPARSER;              use myString_mod
use seism_uti_module;      use datatype0
use myIO
implicit none
private
character(len=100)         :: tmpfnm
public :: construct_types,    destruct_types,    &
          load_n_encode_csg,  randomize_encoder, &
		  test__rand_enc,                        &
		  iter_src_in_bund,   test__iter_src,    &
		  clanRcvrIdxRng_inBun, clanRcvrIdxRng_inSrc
!=========================================
CONTAINS
           !--------------!
	       ! entry points !
	       !--------------!
subroutine  construct_types(e, pr, parfile, existingIt, fname, memsz)
  type(entire_param),   intent(inout)  :: e
  type(proc_type),      intent(inout)  :: pr
  character( len=* ),   intent(in)     :: parfile
  integer,              intent(in)     :: existingIt
  character(len=*),optional,intent(in) :: fname
  real,                 intent(out)    :: memsz
  real                  :: sz1
  character(len=100)    :: fnm
  
  call input_op(e%o, parfile)
  call input_dm(e%d, parfile)
  call input_sc(e%s, parfile)
  call input_io(e%i, parfile)

  call init_entire_param( e, existingIt, memsz )
  call init_pr ( pr, e, sz1 )
  memsz = memsz + sz1
  
!#ifdef DEBUGfwi
  fnm = '';  if(present(fname))      fnm = fname
  call dump_op(e%o,  fnm)
  call dump_dm(e%d,  fnm)
  call dump_sr(e%s,  fnm)
  call dump_io(e%i,  fnm)
  call dump_pr(pr,  e, fnm)
!#endif
end subroutine


subroutine  destruct_types(e, pr)
  type(entire_param),   intent(inout)  :: e
  type(proc_type),      intent(inout)  :: pr
  call kill_entire_param( e )
  call kill_pr( pr, e )
end subroutine


subroutine  load_n_encode_csg(ib, pr, e)! not for freq. selec.
  ! Delay and Polarity encoding, for the current bundle ib
  integer,              intent(in)    :: ib  ! global bundle index
  type(proc_type),      intent(inout) :: pr
  type(entire_param),   intent(in)    :: e
  integer               :: is, sht
  
  if (e%o%SINU>0)       stop 'freq. selec. should not call load_n_encode_csg'
  pr%csg%o    =0.
  if (e%i%csgin_file=='n/a' .or. pr%bi2 < pr%bi1) return  ! no loading
  
  if (e%o%TOW==0)       then    ! fixed spread
    do is = pr%b(ib)%si1, pr%b(ib)%si2
      call filename(tmpfnm, e%i%csgin_file, is, e%i%csgfile_ext)  ! <-- NOTE THE FILENAME
	  call read_binfile(tmpfnm, pr%csg%e)                ! csg%e is temp
	  sht = pr%b(ib)%dly(is)
	  pr%csg%o(1+sht:,:) = pr%csg%o(1+sht:,:) + pr%b(ib)%pol(is) * &
	                       pr%csg%e(1: e%d%nt-sht, :)
    end do
  else                  ! marine, using Ricker implies single source, so bundle & source
    call filename(tmpfnm, e%i%csgin_file, ib,  e%i%csgfile_ext) ! have same global index
    call read_binfile(tmpfnm, pr%csg%o(:,:pr%b(ib)%rj2-pr%b(ib)%rj1+1)) ! no encoding
  end if
end subroutine



subroutine randomize_encoder(pr, e)    ! for all encoders in this proc
  type(proc_type),      intent(inout) :: pr
  type(entire_param),   intent(in)    :: e
  
  integer               :: ib
  real                  :: mn, mx
  integer,  save        :: kit = 1

 do ib = pr%bi1, pr%bi2   ! looping through all bundles
    if(e%s%nbtot  >= e%s%ns) then    ! single source
      pr%b(ib)%pol = 1;  pr%b(ib)%dly = 0
      cycle
	end if  
	call rand_quasi(pr%b(ib)%rndtmp, nsize, rank)
	mn= minval(pr%b(ib)%rndtmp)-1e-5
	mx= maxval(pr%b(ib)%rndtmp)+1e-5
    pr%b(ib)%pol = floor((pr%b(ib)%rndtmp - mn)*(2/(mx - mn)))*2 - 1
	pr%b(ib)%dly = 0
	
	if (e%o%TOW == 0)          then  ! fixed spread
	  if (e%s%MaxSrcDelay > 0) then
	    call rand_quasi(pr%b(ib)%rndtmp, nsize, rank)
	    mn= minval(pr%b(ib)%rndtmp)-1e-5
	    mx= maxval(pr%b(ib)%rndtmp)+1e-5
	    pr%b(ib)%dly = floor( (pr%b(ib)%rndtmp-mn)* (e%s%MaxSrcDelay+1)/(mx-mn) )
	  end if
	else  ! marine towed, multi-src
	  if(e%i%freq_encod_file == 'n/a') then  ! NAIVE RAND_PERM()
        pr%b(ib)%ci_von_nw = rand_unif_spread(pr%b(ib)%nc, e%s%nw) ! to clans per bundle
	  else
		call load_intv_perBun_mpi(pr%b(ib)%ci_von_nw, e%i%freq_encod_file, e%s%nw,&
			             e%s%nbtot, ib, kit)
	  endif
!#ifdef DEBUGfwi
  if(kit==1) then
    write(*,*) 'rank, ib, Encoders: MaxSrcDelay, ...ci_von_nw', &
                rank, ib, e%s%MaxSrcDelay
    ! write(*,'(500I4)') pr%b(ib)%pol
    write(*,'(500I4)') pr%b(ib)%ci_von_nw
    call flush(6)
  end if
!#endif	  
	end if
 end do
 kit= kit + 1
end subroutine

subroutine test__rand_enc(pr, e)
  type(proc_type),      intent(inout) :: pr
  type(entire_param),   intent(in)    :: e
  integer               :: ib
  integer,  save        :: kit = 1
 
 write(*,*) '  test__rand_enc:' 
 do ib = pr%bi1, pr%bi2 
    write(*,*) 'ib,  pol:',   ib 
	write(*,'(100I4)')  pr%b(ib)%pol
	write(*,*) 'dly:'
	write(*,'(100I4)')  pr%b(ib)%dly
	if (e%o%TOW /= 0) then
	  write(*,*) 'ci_von_nw:'
	  write(*,'(100I4)')  pr%b(ib)%ci_von_nw
	end if
	call flush(6)
 end do
 kit= kit + 1
end subroutine

function iterator_depth0(iout, head, tail)    result(conti)
! CAUTION: MAKE SURE THIS IS NOT SHARED BY SEVERAL LOOPS SIMULTANEOUSLY
logical              :: conti
integer,             intent(out) :: iout
integer,             intent(in)  :: head, tail  ! looping range:  [head, tail]
integer,    save     :: i = 0
logical,    save     :: initial = .true.  ! start from (reset to) loop initial
  if (initial)       then
    i = head;        initial = .false.
  end if
  iout    = i
  if (i > tail)      then
    conti = .false.; initial = .true.;  return
  end if
  conti   = .true.
  i       = i + 1
end function
function iterator_depth1(iout, head, tail)    result(conti)
! CAUTION: MAKE SURE THIS IS NOT SHARED BY SEVERAL LOOPS SIMULTANEOUSLY
logical              :: conti
integer,             intent(out) :: iout
integer,             intent(in)  :: head, tail  ! looping range:  [head, tail]
integer,    save     :: i = 0
logical,    save     :: initial = .true.  ! start from (reset to) loop initial
  if (initial)       then
    i = head;        initial = .false.
  end if
  iout    = i
  if (i > tail)      then
    conti = .false.; initial = .true.;  return
  end if
  conti   = .true.
  i       = i + 1
end function

function iter_src_in_bund(is, bn, e, ic, iw) result(conti)
	integer,             intent(out) :: is
	type(bund_type),     intent(in)  :: bn
	type(entire_param),  intent(in)  :: e
	integer, optional,   intent(out) :: ic, iw  ! both present in case of SINU
	logical              :: conti,   cnts
	logical,    save     :: initial = .true. 
	integer,    save     :: iw0=1    ! keeping old iw

	if(e%o%SINU==0)  then !  wavelet in t-domain, no clan trick
	  conti = iterator_depth0(is, bn%si1, bn%si2) 
	  return
	end if
	
	! sinusoid:
	conti = .true.
	if (initial)          then
	  conti   = iterator_depth0(iw0, 1, e%s%nw)
	  initial = .false.
	end if
	ic   = bn%ci_von_nw(iw0)   ! clan local index
	cnts = iterator_depth1(is, bn%c(ic)%si1, bn%c(ic)%si2)

	if(.not. cnts)  then
	  conti   = iterator_depth0(iw0, 1, e%s%nw)
	  if (.not. conti) then
	    initial = .true.;  iw = iw0;    return
	  end if
	  ic   = bn%ci_von_nw(iw0)
	  cnts = iterator_depth1(is, bn%c(ic)%si1, bn%c(ic)%si2)
	end if
	iw = iw0
end function

subroutine  test__iter_src(bn, e)  ! used to debug the iterator()
	type(bund_type),     intent(inout)  :: bn
	type(entire_param),  intent(inout)  :: e
	
	integer :: is, ic, iw
	logical :: conti
		
	if (e%o%SINU==0) then
	  write(*,*) 'si1, si2', bn%si1, bn%si2
	else  
	  write(*,*) 'nc, nw', bn%nc, e%s%nw
	end if
	do while(iter_src_in_bund(is, bn, e, ic, iw))
		write(*,*) 'is, ic, iw', is, ic, iw
		call flush(6)
	end do
	
end subroutine

subroutine   clanRcvrIdxRng_inBun(rng, clan, bun)
! bundle has a working receiver set, which is the union of all its clan's rcvr's
! In this set, index starts at 1.  This set is CSG's dim2  
! Find the index range of the clan's rcvr's in this working set.
integer,         intent(out)   :: rng(:)
type(clan_type), intent(in )   :: clan
type(bund_type), intent(in )   :: bun
rng(1) = clan%rj1 - bun%rj1 + 1   ! rj1's are all global indices, so '-' makes sense
rng(2) = clan%rj2 - bun%rj1 + 1
end subroutine

subroutine   clanRcvrIdxRng_inSrc(rng, clan, s)
! Each source has a stored csgobs (:, s%ri1:s%ri2).
! Among the receivers, find the (global) index range of the clan's rcvr's.
! Recall, clan's aperture is the intersection of all its sources' aperture,
! and therefore no bigger than each individual source's aperture.
integer,         intent(out)   :: rng(:)
type(clan_type), intent(in )   :: clan
type(source_type),intent(in)   :: s

rng(1) = clan%rj1   ! straightforward, since all are consistently global indices
rng(2) = clan%rj2
end subroutine

            !--------------!
    	    ! private func !
	        !--------------!
!=======================================
!
!             INIT & KILL
!
!---------------------------------------
subroutine init_csg(c,  e, ri1, ri2, memsz)
  type(csg_current),    intent(inout) :: c
  type(entire_param),   intent(in)    :: e
  integer,              intent(in)    :: ri1, ri2  ! defining the receiver indices
  real,                 intent(out)   :: memsz
  integer               :: dim1
  
  allocate(   c%c(e%d%nt,  ri1:ri2) )
  memsz = e%d%nt * (ri2-ri1+1) * 4.
  
  if (e%o%SINU == 0)     then 
    allocate( c%o(e%d%nt,  ri1:ri2) )
	memsz = memsz + e%d%nt * (ri2-ri1+1) * 4.
  else  ! else, csg_obs are loaded up for all sources, thus don't need the c%o
    allocate( c%o(0, 0) )
  end if	
  
  dim1 = e%d%nt;         if (e%o%SINU/=0)  dim1 = 2 * e%s%nw  
  
#ifdef DEBUGfwi
     allocate( c%e(  e%d%nt, ri1:ri2 ) )
#else 
  if (e%o%TOW /= 0 .and. e%o%SINU==0 .and. e%o%FWI==0) then
     allocate( c%e(2*e%s%nw, ri1:ri2) )  ! temp. space, for DFT
	 memsz = memsz + 2*e%s%nw * (ri2-ri1+1) * 4.
  else if (e%o%FWI/=0)   then
	 allocate( c%e(   dim1, ri1:ri2) ) 
	 memsz = memsz + dim1 * (ri2-ri1+1) * 4.
  else
    allocate( c%e(0,0) ) 
  end if
#endif  
end subroutine
!.............
subroutine kill_csg(c)
  type(csg_current),     intent(inout) :: c
  deallocate(c%c,  c%o,  c%e)
end subroutine

!--------------------------
subroutine get_spectrum(magn,phas,nw, src_wvlt,trctmp,nt)
! get the abs(.) and phase angle of source wavelet in nt-point DFT
! dw = 2 pi/(dt nt),  nw given by high cutoff freq. of Ricker
! assuming nt >> 2*nw, DC==0 ignored.
real,      intent(out)    ::   magn(:),  phas(:)
integer,   intent(in)     ::   nw,   nt
real,      intent(in)     ::   src_wvlt(:)  ! length nt
real,      intent(out)    ::   trctmp(:)

integer    :: k
real       :: re, im
call rfft1d(src_wvlt, trctmp)
do k       = 1, nw
  re       = trctmp(   1+k)
  im       = trctmp(nt+1-k)
  magn(k)  = sqrt(re*re + im*im)
  phas(k)  = atan2(im,    re)
end do
end subroutine
!..................................

subroutine init_entire_param( e, existingIt, memsz)
  type(entire_param),  intent(inout) :: e
  integer,             intent(in)    :: existingIt
  real   ,             intent(out)   :: memsz
  integer              :: tmp, zhat
  
  e%s%nbtot   =  min(e%s%nbtot, e%s%ns) ! ensuring no bundle will be empty
  
  if (e%o%FWI == 0) then       ! just forward modeling
    if (e%i%csgout_file == 'n/a')  stop 'csgout_file needed for forward modeling!'
	call dir_insert(e%i%csgout_file,  int(e%s%f0))
	
	e%s%nbtot =  e%s%ns         ! enforce single source when generating true CSG
	e%s%ncb   = 1
	if (e%i%vel_tru_file /='n/a') e%i%vel_ini_file = e%i%vel_tru_file
  end if
  
  !!!!!! op:
  e%o%SINU = 1
  if(e%o%TOW==0 .or. e%s%nbtot >= e%s%ns) e%o%SINU = 0 ! fixed spread or single source
  if(e%o%SINU==0 .and. e%o%MUTE ==2)   e%o%MUTE = 0 ! when not sinusoid --> MUTE==2 no sense 
  
  !!!!!! dm:
  if (e%o%FREESURF>1) then
	  e%d%nz = e%d%nz + e%o%FREESURF  ! input v will be extended accordingly
  end if
  e%d%x3 = 1;                  e%d%z3 = 1
  e%d%x4 = e%d%nx;             e%d%z4 = e%d%nz
  e%d%x5 = e%d%x4+1;           e%d%z5 = e%d%z4+1
  e%d%x6 = e%d%x4+e%d%npml;    e%d%z6 = e%d%z4+e%d%npml
  e%d%x2 = e%d%x3-1;           e%d%z2 = e%d%z3-1
  e%d%x1 = e%d%x3-e%d%npml;    e%d%z1 = e%d%z3-e%d%npml
  if (e%o%FREESURF==1)         e%d%z1 = e%d%z3

  
  !!!!!! sr
  call init_fftw(e%d%nt )
  
  e%s%dw = TWOPI/(e%d%nt * e%d%dt)
  e%s%w1 = e%s%dw
  if(e%s%nw<=0)   e%s%nw = int((TRUNC_F_RATIO* e%s%f0)* (e%d%nt* e%d%dt) +.5) !fhi/df
  
    ! source & receiver depth shift, due to adaptive layer trick, possibly:
  zhat   = 0
  if(e%o%FREESURF>1)        zhat = e%o%FREESURF
  e%s%zs = e%s%zs + zhat;   e%s%zg= e%s%zg + zhat
  
    ! symetric aperture
  if (e%o%TOW== 2)          then
    tmp       = e%s%slead + e%s%saptr - 1
	e%s%saptr = 2*tmp + 1
	e%s%slead =  -tmp
  end if
  
    ! window length - 1:
  if(e%o%TOW==0 .and. e%o%MUTE==3)  e%s%mskWinLen1 = ceiling(2.4/e%d%dt/e%s%f0)-1
    ! for phase encoding:
  e%s%MaxSrcDelay = 0
  if (e%o%SINU==0) e%s%MaxSrcDelay= ceiling(0.5/e%d%dt/e%s%f0)
  if (e%s%nbtot >= e%s%ns)          e%s%MaxSrcDelay= 0 ! single source
  
  memsz = 0.
    ! water bottom
  if (e%s%z_fix_v > 0) then     !  else no water layer
	  if (e%i%waterbottom_file /= 'n/a')  then
		allocate( e%s%waterBtm(e%d%nx) )
		if (rank==0) call read_tex(e%s%waterBtm, e%i%waterbottom_file, e%d%nx)
		call bcast(e%s%waterBtm)
		memsz = memsz + e%d%nx * 4.
		e%s%z_fix_v = zhat + 1
	  end if ! else, water layer is [1, z_fix_v]. this ignores non-zero zhat case.
  else
      e%i%waterbottom_file = 'n/a'
	  allocate(e%s%waterBtm(0)) ! dummy 
  end if
    ! source wavelet:
  allocate( e%s%s(e%d%nt), e%s%trctmp(e%d%nt) )  ! in d/dt2, we need a temp trace
  memsz = memsz + e%d%nt *2* 4.                    ! also, fft requires a temp trace
  if (rank==0) then
	  if(e%i%srcwvlt_file /= 'n/a')   then
		e%s%s = 0.
		tmp   = -1
		call read_binfile(e%i%srcwvlt_file, e%s%s, tmp)
	  else
		call ricker_t(    e%s%s,   e%s%f0,  e%d%dt)
	  end if
  end if
  call bcast(e%s%s)
  
  if (e%o%TOW /= 0 .and. e%s%nbtot < e%s%ns) then  ! marine multi-src  --> freq.selec.
    allocate( e%s%sw(e%s%nw),  e%s%phw(e%s%nw) )
	memsz = memsz +  e%s%nw *2  * 4.
	call  get_spectrum(e%s%sw, e%s%phw, e%s%nw,  e%s%s,  e%s%trctmp, e%d%nt)
  end if
  
  
  !!!!!! io:
  call append_BundlesNum(e%i%vel_out_file, e%s%nbtot)
  call append_BundlesNum(e%i%grd_ite_file, e%s%nbtot)
  call append_BundlesNum(e%i%vel_Err_file, e%s%nbtot, '.dat')
  call append_BundlesNum(e%i%dat_Err_file, e%s%nbtot, '.dat')
  
  e%i%V_TRU_AVAIL = e%i%vel_tru_file /= 'n/a' .and. e%i%vel_Err_file /= 'n/a' &
                   .and. e%o%FWI /= 0 .and. rank==0  ! only root handles simple I/O
  ! vel_tru_file is only used to save computed vel_err.
  
  if(e%i%csgin_file /='n/a')     call dir_insert(e%i%csgin_file, int(e%s%f0))
  ! e.g. ./csg15/csgxxx
  
  if(existingIt>0 .and. rank==0) call &
        filename(e%i%vel_ini_file, e%i%vel_out_file, existingIt,'.dat')
  
end subroutine
!.............
subroutine kill_entire_param( e)
  type(entire_param),             intent(inout) :: e
  if(allocated(e%s%waterBtm))     deallocate(e%s%waterBtm   )
  if(allocated(e%s%sw)      )     deallocate(e%s%sw, e%s%phw)
  deallocate(  e%s%s, e%s%trctmp )
end subroutine
!--------------------------

subroutine init_pr(pr,  e, memsz)
  type(proc_type),      intent(inout) :: pr
  type(entire_param),   intent(in)    :: e
  real,                 intent(out)   :: memsz
  integer               :: wid, ib
  real                  :: sz1  ! mem size
  
  call init_random_seed(rank)
  
  ! see what bundles are assigned to process rank out of nsize total processes:
  call partitionBegEnd(pr%bi1, pr%bi2, e%s%nbtot, nsize,   rank,   0, .true.) 
  if (pr%bi1 > pr%bi2)    return  ! no nundles are assigned to this proc
  
  allocate( pr%b(pr%bi1 : pr%bi2) )
  
  wid   = 0
  memsz = 0.
  do ib = pr%bi1,  pr%bi2
    call init_bund(pr%b(ib), ib)
	wid = max(wid, pr%b(ib)%ri2 - pr%b(ib)%ri1 + 1)
  end do

  call init_csg(pr%csg,  e, 1, wid, sz1)    ! as the bundles share a current working
  ! csg, the receivers may be different.  So, we can't attach a physical global
  ! index range to the csg.   What we do is to left aligned the invoked csg for
  ! a specific bundle, which is always no bigger than the allocated size of the 
  ! common csg. 
  memsz = memsz + sz1
  
  ! : : : : : : : : : 
       contains
	   
  subroutine init_bund(b, ib)
    type(bund_type),    intent(inout) :: b
    integer,            intent(in)    :: ib    ! global bundle index
	integer             :: ic, ns
	call partitionBegEnd(b%si1, b%si2, e%s%ns,    e%s%nbtot, ib-1)
	
	ns    = max(0, b%si2 - b%si1 + 1)  ! # sources in this boundle
	b%nc  = min(e%s%ncb, ns)           ! nc cannot exceed ns, so no clan can be empty
	allocate ( b%c(b%nc) )
	do ic = 1, b%nc
	  call init_clan(b%c(ic), ic, b%nc, b%si1, b%si2)
	end do

	b%rj1 = b%c(  1 )%rj1
	b%rj2 = b%c(b%nc)%rj2    ! intersected receivers
	
	b%ri1 = b%c(  1 )%ri1
	b%ri2 = b%c(b%nc)%ri2    ! non-intersec ...
	
	if(e%o%SINU==1)  then
	  allocate( b%ci_von_nw(e%s%nw) )
	  memsz = memsz + e%s%nw *4.
    end if
	allocate( b%pol(b%si1:b%si2), &
	          b%dly(b%si1:b%si2), &
		   b%rndtmp(b%si1:b%si2)  )
	memsz = memsz + 3*(b%si2 - b%si1 + 1)*4.	   
  end subroutine
  
  subroutine init_clan(c, ic, nc, s1, s2)
    type(clan_type),    intent(inout) :: c
    integer,            intent(in)    :: ic, nc, s1, s2
	integer             :: is
	! ic:    local clan index, 1, ..., b%nc
	! s1,s2: clan's father bundle's source global index range
	
	call partitionBegEnd(c%si1, c%si2, s2-s1+1, nc, ic-1, s1-1)
	allocate( c%s(c%si1 : c%si2) )
	do is = c%si1, c%si2
	  call init_src( c%s(is), is )
	end do
	c%ri1 = c%s(c%si1)%ri1
	c%ri2 = c%s(c%si2)%ri2
	
	c%rj1 = c%s(c%si2)%ri1   
	c%rj2 = c%s(c%si1)%ri2    ! intersected receivers
  end subroutine
  
  subroutine init_src( s, is)
    type(source_type),  intent(inout) :: s
    integer,            intent(in)    :: is    ! global src index
	
	if(e%o%TOW == 0)    then
	  s%ri1= 1;         s%ri2= e%s%ng
	else  
	  s%ri1= ridx1(is); s%ri2= ridx2(is)       ! untruncated rcvr idx
	end if
	
	if(e%o%SINU==1)     then
	  allocate( s%csgobs(2*e%s%nw, s%ri1:s%ri2) )
	  memsz = memsz + (2*e%s%nw * (s%ri2 - s%ri1 + 1)) * 4.
	  if (e%i%csgin_file == 'n/a') then   ! must be in debug mode, allow this
	    write(*,*) 'csgin_file inexistent for FWI, SINU !!!' ! issue a warning only
		s%csgobs = 0.
	  else
	    call filename(tmpfnm, e%i%csgin_file, is, e%i%csgfile_ext) ! <--- NOTE THE FILENAME
        call read_binfile( tmpfnm,   s%csgobs )  ! load the full aperture data
      end if
	else
      allocate( s%csgobs(0,0) )	
	end if  
  end subroutine 
  
  ! . . . . . . . . .	   
  integer function ridx1(is) ! find source is's 1st receiver's index
   integer, intent(in) ::is  ! global src idx
   ridx1 = e%s%s1 + (is-1)* e%s%ds + e%s%slead     ! x-coord
   ridx1 = max(1, 1 + ceiling( real(ridx1 - e%s%g1)/real(e%s%dg) ))
  end function
 
  integer function ridx2(is) ! find source is's last receiver's index
   integer, intent(in) ::is
 !  ridx2 = e%s%s1 + (is-1)* e%s%ds + e%s%slead +  (e%s%saptr-1)* e%s%dg
 !  ridx2 = min( e%s%ng, floor( real(ridx2- e%s%g1)/real(e%s%dg) ) + 1 )
   ridx2 = min(e%s%ng, ridx1(is) + e%s%saptr - 1)
  end function
  
end subroutine
!.............

subroutine kill_pr(pr, e)
  type(proc_type),      intent(inout) :: pr
  type(entire_param),   intent(in)    :: e
  integer   ib
  if (pr%bi1 > pr%bi2 )  return  ! this proc is empty, no bundles assigned
  
  do ib = pr%bi1, pr%bi2
	call kill_bund(pr%b(ib))
  end do
  
  deallocate(   pr%b  )
  call kill_csg(pr%csg)
  ! : : : : : : : : :
       contains
	   
  subroutine  kill_bund(b)
    type(bund_type),   intent(inout) :: b
	
    integer   ic
	do ic = 1, b%nc
	  call kill_clan(b%c(ic ))
	end do
	
	if(allocated(b%pol))        deallocate(b%pol, b%dly, b%rndtmp)
	if(allocated(b%ci_von_nw))  deallocate(b%ci_von_nw           )
	deallocate(  b%c   )
  end subroutine
  
  subroutine  kill_clan(c)
    type(clan_type),   intent(inout) :: c
    integer   is
	do is = c%si1, c%si2
	  call kill_src( c%s(is) )
	end do
	deallocate( c%s )
  end subroutine
  
  subroutine  kill_src( s)
    type(source_type),  intent(inout) :: s
    deallocate( s%csgobs )
  end subroutine  
end subroutine
!=======================================
!
!             READ
!
!---------------------------------------
subroutine input_op(o, parfile)
  type(option_param),  intent(inout) :: o
  character( len=*),   intent(in)    :: parfile
  integer(1)           ::     ii
  if (rank == 0)       then
  	  call readParFile(parfile,'CGreset',       o%CGreset,      3)
	  call readParFile(parfile,'CGresetBlock',  o%CGresetBlock, 4)
	  call readParFile(parfile,'FREESURF',      o%FREESURF,     0)  ! FREESURF
	  ii=o%FREESURF;  call readCmd('FREESURF',  o%FREESURF,    ii)  !    read von command line, overriding parfile
	  call readParFile(parfile,'FWI',           o%FWI,          1)
	  ii=o%FWI;       call readCmd('FWI',       o%FWI,         ii)  ! FWI,    von command line
	  call readParFile(parfile,'MUTE',          o%MUTE,         1)
	  
  	  call readParFile(parfile,'TOW',           o%TOW,          1)
	  ii=o%TOW;       call readCmd('TOW',       o%TOW,         ii)  ! TOW,    von command line
	  call readParFile(parfile,'PRECND',        o%PRECND,       2)
	  ii=o%PRECND;    call readCmd('PRECND',    o%PRECND,      ii)  ! PRECND, von command line
  	  call readParFile(parfile,'FWD_sbsmp',     o%FWD_sbsmp,    1)
	  call readParFile(parfile,'Derr_mode',     o%Derr_mode,    1)
  end if
  call MPI_BCAST(o%CGreset,        1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%CGresetBlock,   1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%FREESURF,       1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%FWI,            1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%MUTE,           1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  
  call MPI_BCAST(o%TOW,            1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%PRECND,         1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  
  call MPI_BCAST(o%FWD_sbsmp,      1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(o%Derr_mode,      1,MPI_INTEGER1, 0,MPI_COMM_WORLD,ierr)
end subroutine
!-------------
subroutine input_dm(m, parfile)
  type(domain_param),  intent(inout) :: m
  character( len=*),   intent(in)    :: parfile
  if (rank == 0)       then
	  call readParFile(parfile,'FDhalfw',       m%lx,     4)
	  call readParFile(parfile,'npml',          m%npml,  25)
	  call readParFile(parfile,'nz',            m%nz)
	  call readParFile(parfile,'nx',            m%nx)
	  call readParFile(parfile,'nt',            m%nt) 

	  call readParFile(parfile,'dx',            m%dx)
	  call readParFile(parfile,'dt',            m%dt)
	  call readParFile(parfile,'vmax',          m%vmax, -1.)
	  if (m%vmax<0.)  call readParFile(parfile,'vmx',m%vmax)
	  call readParFile(parfile,'vmin',          m%vmin, -1.)
	  if (m%vmin<0.)  call readParFile(parfile,'vmn',m%vmin)
  endif
  call MPI_BCAST(m%lx,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)  
  call MPI_BCAST(m%npml,           1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%nz,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%nx,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%nt,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)

  call MPI_BCAST(m%dx,             1,MPI_REAL,     0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%dt,             1,MPI_REAL,     0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%vmax,           1,MPI_REAL,     0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(m%vmin,           1,MPI_REAL,     0,MPI_COMM_WORLD,ierr)
end subroutine
!-------------
subroutine input_sc(a, parfile)
  type(source_param),  intent(inout) :: a
  character( len=*),   intent(in)    :: parfile
  integer              :: ii
  if (rank == 0)       then
      call readParFile(parfile,'nw',               a%nw,   -1)
      call readParFile(parfile,'nbundles',         a%nbtot, 1)
	  ii = a%nbtot;    call readCmd('nbundles',    a%nbtot,ii)  ! von cmd line
      call readParFile(parfile,'nclans_per_bundle',a%ncb,   1)
	  
	  call readParFile(parfile,'f0',            a%f0)
	  call readParFile(parfile,'s1',            a%s1,   1)
	  call readParFile(parfile,'ds',            a%ds)
	  call readParFile(parfile,'ns',            a%ns)
	  call readParFile(parfile,'zs',            a%zs,   1)
	  
	  call readParFile(parfile,'g1',            a%g1,   1)
	  call readParFile(parfile,'dg',            a%dg)
	  call readParFile(parfile,'ng',            a%ng)
	  call readParFile(parfile,'zg',            a%zg,   2)
	  call readParFile(parfile,'z_fix_v',       a%z_fix_v, -1) ! no water layer
	  
	  call readParFile(parfile,'slead',         a%slead,-1)
	  call readParFile(parfile,'saptr',         a%saptr,-1)
  end if
  call MPI_BCAST(a%nw,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%nbtot,          1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%ncb,            1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  
  call MPI_BCAST(a%f0,             1,MPI_REAL,     0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%s1,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%ds,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%ns,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%zs,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  
  call MPI_BCAST(a%g1,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%dg,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%ng,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%zg,             1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%z_fix_v,        1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  
  call MPI_BCAST(a%slead,          1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(a%saptr,          1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
end subroutine
!-------------
subroutine input_io(i, parfile)
  type(io_param_typ),  intent(inout) :: i
  character( len=*),   intent(in)    :: parfile
  ! if a file's I/O is potentially used by all MPI processes, then naive handling will
  ! lead to conflict.  In this case, only let the root do I/O and then broadcast file
  ! content (if I) after reading.  If each process deals w/ its own file, like CSG, 
  ! travelTime (which pertain to individual source, and therefore would be accessed by 
  ! one MPI processes at most), then all processes do I/O and thus need the filenames.
  ! Waterbottom_file  is needed by all for making judgement.
  if (rank == 0)       then
	  call readParFile(parfile,'XFAST',           i%XFAST,      .false.) ! no broadcast
	  call readParFile(parfile,'vel_ini_file',    i%vel_ini_file, 'n/a') ! no broadcast
	  if (i%vel_ini_file=='n/a') call readParFile(parfile,'vel_file', i%vel_ini_file)
	  call readParFile(parfile,'vel_out_file',    i%vel_out_file, 'n/a') ! no broadcast
	  call readParFile(parfile,'vel_tru_file',    i%vel_tru_file, 'n/a') ! no broadcast
	  
	  call readParFile(parfile,'vel_Err_file',    i%vel_Err_file, 'n/a') ! no broadcast
	  call readParFile(parfile,'dat_Err_file',    i%dat_Err_file, 'n/a') ! no broadcast
	  call readParFile(parfile,'grd_ite_file',    i%grd_ite_file, 'n/a') ! no broadcast

	  call readParFile(parfile,'csgin_file',      i%csgin_file,   'n/a')
	  call readParFile(parfile,'csgout_file',     i%csgout_file,  'n/a')
	  call readParFile(parfile,'csgfile_ext',     i%csgfile_ext,  '.bin')
	  call readParFile(parfile,'trvlTim_file',    i%trvlTim_file, 'n/a')
	  call readParFile(parfile,'waterbottom_file',i%waterbottom_file,'n/a')
	  call readParFile(parfile,'srcwvlt_file',    i%srcwvlt_file,    'n/a')! no brdcst 
	  call readParFile(parfile,'freq_encod_file', i%freq_encod_file, 'n/a')
	  tmpfnm= i%freq_encod_file
	  call readCmd(            'freq_encod_file', i%freq_encod_file, tmpfnm)! von cmd line
  end if
  call MPI_BCAST(i%csgin_file,      100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(i%csgout_file,     100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(i%csgfile_ext,     100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(i%trvlTim_file,    100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(i%waterbottom_file,100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(i%freq_encod_file, 100,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
end subroutine
!=======================================
!
!             PRINT
!
!---------------------------------------

subroutine dump_op(op,   fname)
  type(option_param),    intent(in) :: op
  character(len=*),      intent(in),   optional :: fname
  character(len=100)     fnm
  if (rank/=nsize-1)     then  
    call MPI_barrier(MPI_COMM_WORLD, ierr); return
  end if
  fnm = '';  if(present(fname))      fnm =  fname

  call write_tex(fnm, '/CGreset, CGresetBlock, FREESURF, FWI:'  )
  call write_tex(fnm, (/op%CGreset,op%CGresetBlock,op%FREESURF,op%FWI/))

  call write_tex(fnm, '/MUTE,   SINU,   TOW,    PRECND:'     )
  call write_tex(fnm, (/op%MUTE, op%SINU, op%TOW, op%PRECND/))
  
  call write_tex(fnm, '/FWD_sbsmp,  Derr_mode:'        )
  call write_tex(fnm, (/op%FWD_sbsmp, op%Derr_mode   /))
  
  call MPI_barrier(MPI_COMM_WORLD, ierr)
end subroutine

subroutine dump_dm(dm,   fname)
  type(domain_param),    intent(in) :: dm
  character(len=*),      intent(in),   optional :: fname
  character(len=100)     fnm
  if (rank/=nsize-1)     then  
    call MPI_barrier(MPI_COMM_WORLD, ierr); return
  end if
  fnm = '';  if(present(fname))      fnm  = fname

  call write_tex(fnm, '/lx,   npml,   nz,    nx,    nt:'          )
  call write_tex(fnm, (/dm%lx, dm%npml, dm%nz, dm%nx, dm%nt     /))

  call write_tex(fnm, '/z1,   z2,    z3,   z4,   z5,   z6:'       )
  call write_tex(fnm, (/dm%z1, dm%z2, dm%z3, dm%z4, dm%z5, dm%z6/))
  
  call write_tex(fnm, '/x1,   x2,    x3,   x4,   x5,   x6:'       )
  call write_tex(fnm, (/dm%x1, dm%x2, dm%x3, dm%x4, dm%x5, dm%x6/))  
  
  call write_tex(fnm, '/dx,    dt,    vmax,   vmin:'              )
  call write_tex(fnm, (/dm%dx, dm%dt, dm%vmax, dm%vmin          /))
  
  call MPI_barrier(MPI_COMM_WORLD, ierr)
end subroutine

subroutine dump_sr(sr,   fname)
  type(source_param),    intent(in) :: sr
  character(len=*),      intent(in),   optional :: fname
  character(len=100)     fnm
  if (rank/=nsize-1)     then  
    call MPI_barrier(MPI_COMM_WORLD, ierr); return
  end if
  fnm = '';  if(present(fname))      fnm =  fname

  call write_tex(fnm, '/nw,  nbtot, ncb:'                )
  call write_tex(fnm, (/sr%nw, sr%nbtot, sr%ncb        /))

  call write_tex(fnm, '/w1,  dw,   f0:'                  )
  call write_tex(fnm, (/sr%w1, sr%dw,  sr%f0           /))
  
  call write_tex(fnm, '/s1,  ds,   ns,  zs:'             )
  call write_tex(fnm, (/sr%s1, sr%ds,  sr%ns, sr%zs    /))  
  
  call write_tex(fnm, '/g1,  dg,   ng,  zg:'             )
  call write_tex(fnm, (/sr%g1, sr%dg,  sr%ng, sr%zg    /))
  
  call write_tex(fnm, '/slead, saptr, z_fix_v:'         )
  call write_tex(fnm, (/sr%slead, sr%saptr, sr%z_fix_v/))
  
  call write_tex(fnm, '/mskWinLen1, MaxSrcDelay:'        )
  call write_tex(fnm, (/sr%mskWinLen1, sr%MaxSrcDelay  /))  
  
  if (allocated(sr%waterBtm)) then
    call write_tex(fnm, '/size(waterBtm):'    )
    call write_tex(fnm, (/size(sr%waterBtm) /))
  end if
  
  call write_tex(fnm, '/s'          )
  call write_tex(fnm, sr%s          )  
  
  if (allocated(sr%sw)) then
	  call write_tex(fnm, '/sw'     )
	  call write_tex(fnm, sr%sw     )  

	  call write_tex(fnm, '/phw'    )
	  call write_tex(fnm, sr%phw    ) 
	  
	  call write_tex(fnm, '/size(trctmp)')
	  call write_tex(fnm, size(sr%trctmp))  
  end if
  call MPI_barrier(MPI_COMM_WORLD, ierr  )
end subroutine

subroutine dump_io(io,   fname)
  type(io_param_typ),    intent(in) :: io
  character(len=*),      intent(in),   optional :: fname
  character(len=100)     fnm
  if (rank/=0)           then  
    call MPI_barrier(MPI_COMM_WORLD, ierr);       return
  end if
  fnm = '';  if(present(fname)) fnm = fname
  
  call write_tex(fnm, '/V_TRU_AVAIL,  XFAST:'         )
  call write_tex(fnm, (/io%V_TRU_AVAIL, io%XFAST/)    )
  
  call write_tex(fnm, 'vel_ini_file='//io%vel_ini_file)
  call write_tex(fnm, 'vel_out_file='//io%vel_out_file) 
  call write_tex(fnm, 'vel_tru_file='//io%vel_tru_file) 
  
  call write_tex(fnm, 'vel_Err_file='//io%vel_Err_file)
  call write_tex(fnm, 'dat_Err_file='//io%dat_Err_file)
  call write_tex(fnm, 'grd_ite_file='//io%grd_ite_file)
  
  call write_tex(fnm, 'csgin_file  ='//io%csgin_file  )
  call write_tex(fnm, 'csgout_file ='//io%csgout_file )
  call write_tex(fnm, 'csgfile_ext ='//io%csgfile_ext )
  call write_tex(fnm, 'trvlTim_file='//io%trvlTim_file)  
  
  call write_tex(fnm, 'waterbottom_file='//io%waterbottom_file)
  call write_tex(fnm, 'srcwvlt_file    ='//io%srcwvlt_file    )
  call write_tex(fnm, 'freq_encod_file ='//io%freq_encod_file )    
  
  call MPI_barrier(MPI_COMM_WORLD, ierr)
end subroutine

subroutine dump_pr(pr,   e,  fname)
  type(proc_type),       intent(in)    :: pr
  type(entire_param),    intent(in)    :: e
  character(len=*),      intent(in),   optional :: fname
  character(len=100)     fnm
  integer                :: ib
  
  fnm = ''
  if(present(fname)) then
    fnm = fname
    call filename_insert(fnm,'',rank)
  end if
  call write_tex(fnm, '/rank,  bi1, bi2 (on this PROC):' )
  call write_tex(fnm, (/rank, pr%bi1, pr%bi2/) )
  if (pr%bi1 > pr%bi2)   return
  
  call write_tex(fnm, '/shapes of csg%c, o, e:')
  call write_tex(fnm, (/shape(pr%csg%c), shape(pr%csg%o), shape(pr%csg%e)/) )
  do ib = pr%bi1, pr%bi2
    call dump_bu(pr%b(ib), fnm, ib)
  end do

      contains
  
	subroutine dump_bu(b,    fnm,   ib)
	  type(bund_type),       intent(in) :: b
	  character(len=*),      intent(in) :: fnm
	  integer,               intent(in) :: ib  ! global bundle index
	  
	  integer                :: ic
	  call write_tex(fnm, '/BUND_id, nc, [si1, si2]:'     )
	  call write_tex(fnm, (/ ib, b%nc, b%si1, b%si2     /))
	  call write_tex(fnm, '...[ri1, ri2], [rj1, rj2]:'    )
	  call write_tex(fnm, (/ b%ri1, b%ri2, b%rj1, b%rj2 /))
	  
	  do ic = 1, b%nc
		 call dump_cl(b%c(ic), fnm, ic, ib)
	  end do
	end subroutine

	subroutine dump_cl(c,    fnm,  ic,  ib)
	  type(clan_type),       intent(in) :: c
	  character(len=*),      intent(in) :: fnm
	  integer,               intent(in) :: ic  ! local  clan index
	  integer,               intent(in) :: ib  ! global bundle index 
	  
	  integer                :: is
	  call write_tex(fnm, '/bund_id, CLAN_id, [si1, si2], [ri1, ri2], [rj1, rj2]:')
	  call write_tex(fnm, (/ib,   ic,   c%si1, c%si2, c%ri1, c%ri2, c%rj1, c%rj2/))
	  do is = c%si1, c%si2
		 call dump_sr(c%s(is), fnm, is)
	  end do
	end subroutine

	subroutine dump_sr(s,    fnm,   is)
	  type(source_type),     intent(in) :: s
	  character(len=*),      intent(in) :: fnm
	  integer,               intent(in) :: is  ! global src index
	  
	  call write_tex(fnm, '/SRC_id, ri1, ri2 // shape(csgobs):')
	  call write_tex(fnm, (/is, s%ri1, s%ri2/) )
	  call write_tex(fnm, shape(s%csgobs)      )
	end subroutine
end subroutine

END MODULE
