!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     I/O
!-----------------------


MODULE myIO
use nrtype;       use myPARSER;    use myString_mod
use datatype0;    use myMPI0;      use seism_uti_module

implicit none
! geared toward marine multisource w/ frequency selection
character(len=100) :: tmpfnm
private               tmpfnm
contains
!===================================
subroutine read_parfile(parfile, maxIt, existingIt)
character( len=*),  intent(in)   ::  parfile
integer,            intent(out)  ::  maxIt, existingIt

if (rank == 0) then
  call readParFile(parfile,'maxIter',  maxIt,        20)
  call readParFile(parfile,'extIter',  existingIt,    0)    ! existing # Iter's  
end if  ! Broadcast input parameters:
call MPI_BCAST(maxIt,            1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
call MPI_BCAST(existingIt,       1,MPI_INTEGER,  0,MPI_COMM_WORLD,ierr)
end subroutine
!-----------------------------------
subroutine readPickedTravTime_FixSpread(tim, dt, ng, ns, fname)
integer,          intent(out)   :: tim(:,:)   ! ng x ns  trav tim (in time steps)
real,             intent(in)    :: dt
integer, 		  intent(in)    :: ng, ns
character(len=*), intent(in)    :: fname
integer           :: is, ig
real              :: t
if(fname == 'n/a') then          ! no traveltime file, no reading
  tim =  1
  return
end if
tim   = -1                       ! <=0 entries will be ignored from fwi misfit
open(10, file=fname, form='formatted')
do 
  read(10,*,end=99) is, ig, t    ! each row of the file is arranged as such
  if (is> ns .or. ig> ng) cycle
  tim(ig, is)  = 1 + floor(t/dt) ! '1' because Fortran array index: 1,...
end do
99  continue
close(10)
end subroutine
!-----------------------------------
subroutine flood_up(v, npad)
real,             intent(inout) :: v(:,:)
integer,          intent(in)    :: npad
integer           :: iz
do iz = 1, npad
  v(:,iz) = v(:,npad+1)
end do
end subroutine
!-----------------------------------
subroutine IO_model(v, velfile, xfast, is_in, nzdisk, z1mem)
! to allow the case:  velocity will be extended in z than what's read from disk
real,             intent(inout) :: v(:,:) ! in v, x is fast dim.
character(*),     intent(in)    :: velfile
logical,          intent(in)    :: xfast  ! x is the fast dim on disk
logical,          intent(in)    :: is_in  ! else, it's out
integer, optional,intent(in)    :: nzdisk, z1mem
! nzdisk:   the length of z in disk file, could be smaller than v's z dim.
! z1mem:    when load to v(:,:), put in v(:, z1mem:...), i.e., the starting index
integer           :: ix, iz, nx, nzd, z1

if (rank /=0)       return      !  only the root calls this routine
nx = size(v,1)
if (present(z1mem)) then
  nzd = nzdisk
  z1  = z1mem
else
  nzd = size(v,2)
  z1  = 1
end if

if (xfast) then
  open(10, file=velfile,access='direct', err=20, recl=i4*nx)
  do iz= z1, z1 + nzd - 1
	if (is_in) then 
	  read( 10,rec=iz) v(:, iz)
	else
	  write(10,rec=iz) v(:, iz)
	end if
  enddo
else
  open(10, file=velfile,access='direct', err=20, recl=i4*nzd)
  do ix=1, nx
	if (is_in) then 
	  read( 10,rec=ix) v(ix, z1: z1+nzd-1)
	else
	  write(10,rec=ix) v(ix, z1: z1+nzd-1)
	end if
  enddo
end if
close(10)
return
20   stop '  IO_model error!'
end subroutine
!-----------------------------------
subroutine append_BundlesNum(fnm,  nbtot,  Ext)
 character(*),    intent(inout) :: fnm
 integer,         intent(in)    :: nbtot
 character(*),    intent(in),   optional :: Ext
 
 if (fnm =='n/a') return
 if(present(Ext)) then
     fnm = trim(fnm) // 'B' // trim(i2a(nbtot)) // trim(Ext)
 else
     fnm = trim(fnm) // 'B' // trim(i2a(nbtot)) //'_'
 end if  
end subroutine
!-----------------------------------
subroutine save_stuff(verr, misfit, v, it, iopar, op, v_tru, grd) 
! saving stuff per iter, meant for RANK==0
! v_tru must at least have dummy space for this argument passing to work
real,              intent(out)   :: verr
real,              intent(in)    :: misfit, v_tru(:,:)
real,              intent(inout) :: v(:,:)
integer,           intent(in)    :: it
type(io_param_typ),intent(in)    :: iopar
type(option_param),intent(inout) :: op
real,optional,     intent(inout) :: grd(:,:)

logical            :: is_in

verr = 0.
if (rank /=0)      return        !  only the root calls this routine
if (iopar%dat_Err_file /= 'n/a')  then
  verr    = sqrt(misfit)             ! down to REAL, tmp var here
  call write_binfile(iopar%dat_Err_file, verr, -1)
end if

if (iopar%V_TRU_AVAIL)            then
  if (op%FREESURF>1)               then 
	verr = vel_Err(v(:, op%FREESURF+1:), v_tru)
  else
	verr = vel_Err(v, v_tru)      ! down to REAL
  end if
  call write_binfile(iopar%vel_Err_file, verr, -1)
else
  verr   = -1.  ! denote invalid
end if

is_in    = .false.
if (iopar%vel_out_file /= 'n/a') then
  call filename(tmpfnm, iopar%vel_out_file, it, '.dat')
  call IO_model(v,   tmpfnm, iopar%XFAST, is_in)
end if

if (iopar%grd_ite_file /= 'n/a' .and. present(grd)) then
  call filename(tmpfnm, iopar%grd_ite_file, it, '.dat')
  call IO_model(grd, tmpfnm, iopar%XFAST, is_in)
end if

if (verr>0.) then
	write(*,298)  it, sqrt(misfit), verr
else
	write(*,299)  it, sqrt(misfit)
end if
298 format('Iter, WvfrmErr, VeloErr', I4,       2ES13.4)
299 format('Iter, WvfrmErr         ', I4,        ES13.4)
call flush(6)

end subroutine
!-----------------------------------
real(DP) function vel_Err(v, v_tru)                     ! for rank==0 only
real,              intent(in)    :: v(:,:), v_tru(:,:)  ! same size
integer            :: shp(2),    ix,    iz

if (rank /=0)      return
shp = shape(v)
vel_Err  = 0.
!$OMP PARALLEL DO PRIVATE(ix) REDUCTION(+:vel_Err)
do iz    = 1, shp(2)
  do ix  = 1, shp(1)
    vel_Err = vel_Err + (v(ix, iz) - v_tru(ix, iz))**2
  end do
end do
!$OMP END PARALLEL DO
vel_Err  = sqrt(vel_Err)
end function
!.....................................
subroutine load_intv_perBun_mpi(A, fname, nw, nbtot, bi, kit)
! integer array of size  nw x (nbtot * niter) stored in disk.
! this proc's bundle w/ global index bi; read its column.
! current iter. = kit
integer,           intent(out)   :: A(:)
CHARACTER(LEN=*),  intent(in)    :: fname
integer,           intent(in)    :: nw, nbtot, bi, kit
INTEGER(KIND=MPI_OFFSET_KIND)    :: offset
integer            :: fh, intbytes, count, status(MPI_STATUS_SIZE)

call MPI_FILE_OPEN(MPI_COMM_WORLD, fname, &
                   MPI_MODE_RDONLY, MPI_INFO_NULL, fh, ierr)
call dbg_out
intbytes = kind(fh)
offset   = ((kit-1)*nw*nbtot + (bi-1)*nw) * intbytes
count    = nw
call MPI_FILE_READ_AT(fh, offset, A, count, MPI_INTEGER, status, ierr)
call dbg_out
call MPI_FILE_CLOSE(FH, IERR)
!-------
contains
 subroutine dbg_out
   if (ierr == 0)   return
   write(*,*) 'fname, shape(Buf), ierr', trim(fname), shape(A), ierr
   call flush(6)
   stop
 end subroutine		
end subroutine
!.....................................
END MODULE
