!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     Multi-source
!-----------------------


MODULE source_multi
use myString_mod;   use myRand_module;  use seism_uti_module
use myMPI0;         use myFFTW0;        use datatype0
use datatype1
implicit none
contains

!=================================
subroutine convert2Fdomain(csgf,csgt, trctmp, nt, nw, ng) 
! convert ng time-traces to freq. domain
real,    intent(out) :: csgf(:,:) ! (real, imag) pairs, col. length 2 nw
real,    intent(in)  :: csgt(:,:) ! col. length nt
real,    intent(out) :: trctmp(:) !             nt
integer, intent(in)  :: nt, nw, ng
integer              :: k
integer,    save     :: nw2=0, nw1=0, nw_=0
if(nw2==0)  then
  nw2  = nw*2
  nw1  = nw+1
  nw_  = nt-nw+1
end if
do k = 1, ng
  call rfft1d(csgt(1:nt,k), trctmp)
  csgf(1:nw2:2, k) = trctmp(2:nw1)     ! real
  csgf(2:nw2:2, k) = trctmp(nt:nw_:-1) ! imag
end do
end subroutine
!..................................

subroutine inject_sources(p, xh,xt,zh,zt, tt, bn, e, vdt2, NEG)
integer,             intent(in)   ::   xh,xt, zh,zt
real,                intent(inout):: p(xh:xt, zh:zt) !  p: pressure time slice
integer,             intent(in)   :: tt              ! tt: iter step
type(bund_type),     intent(in)   :: bn
type(entire_param),  intent(in)   :: e
real,                intent(in)   :: vdt2(e%d%x1 : e%d%x6, e%d%z1 : e%d%z6)
logical,optional,    intent(in)   :: NEG  ! if present, must negate the
! injected source, i.e., remove the previously injected source.
integer              :: is,  iw,  ic, sx,  sht, flip
real                 :: w

flip = 1;        if(present(NEG))  flip = -1
if(e%o%SINU==0)  then   !  wavelet in t-domain, no clan trick
  do while(iter_src_in_bund(is, bn, e))
    sx = e%s%s1 + (is-1)*e%s%ds ! p must be big enough to contain all source locations
    sht= tt- bn%dly(is)
    if (sht <= 0)  cycle
    p(sx,e%s%zs) = p(sx,e%s%zs) + flip* vdt2(sx,e%s%zs) * e%s%s(sht) * bn%pol(is)
  end do
else ! sinusoid
  do while(iter_src_in_bund(is, bn, e, iw=iw, ic=ic))
    w  = e%s%w1 + (iw-1)*e%s%dw
    sx = e%s%s1 + (is-1)*e%s%ds
    p(sx,e%s%zs) = p(sx, e%s%zs) + flip* vdt2(sx, e%s%zs) * &
         e%s%sw(iw) * bn%pol(is) * cos( w*((tt-1)*e%d%dt) + e%s%phw(iw) )
  end do
end if
end subroutine
!.....................................
subroutine harvest_rcvrs(csg, p, tt, bn, e) 
type(entire_param), intent(in)   ::   e
! receivers read the data for a bundle ib
real,   intent(out) :: csg(:,:)   !   nt x # receivers
real,   intent(in)  :: p(e%d%x1-e%d%lx: e%d%x6+e%d%lx, e%d%z1-e%d%lx: e%d%z6+e%d%lx)
integer,intent(in)  :: tt         !   tt: iteration step
type(bund_type),    intent(in)   ::   bn

integer :: t2,  g1x,  g2x
t2 = tt;   if ( e%o%MUTE==2 )  t2 = t2 - e%d%nt
if (t2 <= 0)    return
#ifdef DEBUGfwi
 if(t2>e%d%nt) then
   write(*,*) 'tt, MUTE, e%d%nt, t2, shape(csg)', &
               tt, e%o%MUTE, e%d%nt, t2, shape(csg)
   call flush(6)
 end if
#endif
g1x= e%s%g1 + ( bn%rj1 -1 )*e%s%dg
g2x= e%s%g1 + ( bn%rj2 -1 )*e%s%dg
csg(t2, :bn%rj2 - bn%rj1 + 1) = p(g1x: g2x: e%s%dg,  e%s%zg) 
! csg: left-justified, associated to a bundle, may be clipped due to clan
end subroutine
!.....................................
subroutine misfit_adaptGain_F(mis, csg, b, e)
! csg%e stores csg_calcu for the bundle b in f-domain.
! Compute the  error  w.r.t.  csg_obs:  min_{alpha} ||csge||^2,
! where csge = alpha * csg_calcu - csg_obs
!       mis  = ||csge||^2
! all observed data have been kept for each source
! Only deals with relevant subset of sources and frequencies.
! Finally, csge is expressed in (amplitude & phase), ...
real,                 intent(out)   :: mis
type(csg_current),    intent(inout) :: csg
type(bund_type),      intent(in)    :: b
type(entire_param),   intent(in)    :: e

integer               :: f2,   is,  ic,  r,  rngB(2), rngS(2)
real                  :: alpha, nm2,  abstmp
#ifdef DEBUGfwi
  integer,   save     :: kit = 1
#endif

if(e%o%Derr_mode==1) then
	! 1. compute inner prod and norm2:
	alpha   = 0.;      nm2 = 0.
	do while(iter_src_in_bund(is, b, e, iw=f2, ic=ic))
	  f2    = f2 * 2
	  call    clanRcvrIdxRng_inBun(rngB, b%c(ic), b)  ! clan's aperture in csg's dim2
	  nm2   = nm2 + norm2( csg%e( f2-1:f2, rngB(1):rngB(2) ) )
	  
	  call    clanRcvrIdxRng_inSrc(rngS, b%c(ic), b%c(ic)%s(is))
	  alpha = alpha + dot_prod( csg%e( f2-1:f2, rngB(1):rngB(2)), b%pol(is) *    &
	              b%c(ic)%s(is)%csgobs(f2-1:f2, rngS(1):rngS(2)) )
	end do
	alpha   = alpha/nm2
else
  alpha = 1.
end if

! 2. do subtraction
mis     = 0.
#ifdef DEBUGfwi
 if(kit==1) then
   write(*,*) 'b%si1, b%si2', b%si1, b%si2
   write(*,*) 'b%ri1, b%ri2', b%ri1, b%ri2
   write(*,*) 'shape: csg%e', shape(csg%e)
   call flush(6)
 end if
#endif

do while(iter_src_in_bund(is, b, e, iw=f2, ic=ic))
  f2    = f2 * 2
  call    clanRcvrIdxRng_inBun(rngB, b%c(ic), b)
  call    clanRcvrIdxRng_inSrc(rngS, b%c(ic), b%c(ic)%s(is))
#ifdef DEBUGfwi
 if(kit==1) then  
   write(*,'(A, 3I4)') 'is, iw, ic',  is, f2/2, ic
   write(*,'(A, 4I4)') 'rngB, rngS',  rngB, rngS
   write(*,*) 'shape: b%c(ic)%s(is)%csgobs', shape(b%c(ic)%s(is)%csgobs)
   call flush(6)
 end if   
#endif
  csg%e(f2-1:f2, rngB(1):rngB(2)) = alpha* csg%e(f2-1:f2, rngB(1):rngB(2)) - &
                b%pol(is) * b%c(ic)%s(is)%csgobs(f2-1:f2, rngS(1):rngS(2))

  do r  = rngB(1), rngB(2)
	abstmp       =       csg%e(f2,r)**2 + csg%e(f2-1,r)**2
    csg%e(f2,  r)= atan2(csg%e(f2,r),     csg%e(f2-1,r))
    csg%e(f2-1,r)= sqrt( abstmp )
    mis = mis + abstmp
  end do
end do
#ifdef DEBUGfwi
   if(kit==1) kit = 2
#endif 
end subroutine
!.....................................
function compute_misfit(csg, b, e) result(mis)
! csg%c - csg%o   --> csg%e
real                  mis
type(csg_current),    intent(inout) :: csg
type(bund_type),      intent(in)    :: b
type(entire_param),   intent(inout) :: e
integer               :: ng
if (e%o%SINU /= 0)    then
  ! only focusing on the misfit of those assigned freq., while ignoring the rest
  ! csg%c --FT--> csg%e
  call  convert2Fdomain(csg%e, csg%c, e%s%trctmp, e%d%nt, e%s%nw, b%rj2-b%rj1+1) 
  csg%e= csg%e * (2. /real(e%d%nt))  
  ! ^ refer to p.54 of my brown labbook.  --> correct misfit
  call misfit_adaptGain_F(mis, csg, b, e)
else
  ng = b%rj2-b%rj1+1
  if (e%o%Derr_mode==1) then
    call misfit_adaptGain(csg%e, mis, csg%c, csg%o, ng)
  else  
    csg%e(:,:ng) = csg%c(:,:ng)- csg%o(:,:ng)
    mis   =  norm2(csg%e(:,:ng))
  end if
end if
end function
!.....................................
subroutine inject_misfit(p, tt, bn, csge, e, vdt2)   ! p: pressure time slice
type(entire_param),  intent(in)   ::  e
real,   intent(inout):: p(e%d%x1-e%d%lx:e%d%x6+e%d%lx, e%d%z1-e%d%lx:e%d%z6+e%d%lx)
integer,intent(in)   :: tt         ! tt: iteration time step
type(bund_type),     intent(in)   ::  bn
real,                intent(in)   ::  csge(:,:), vdt2(e%d%x1:e%d%x6, e%d%z1:e%d%z6)

integer              :: g1x, g2x, f, f2, is, ic, rngB(2)
real                 :: w

if(e%o%SINU==0)  then   !  wavelet in t-domain--->{fixed spread or single source}
  g1x = e%s%g1 + e%s%dg* (bn%rj1 -1)
  g2x = e%s%g1 + e%s%dg* (bn%rj2 -1)
     p(g1x :g2x: e%s%dg, e%s%zg) =             &
	 p(g1x :g2x: e%s%dg, e%s%zg) +             &
  vdt2(g1x :g2x: e%s%dg, e%s%zg) * csge(tt, :bn%rj2-bn%rj1+1)
  
else                 ! sinusoid, freq. divi.
  do while(iter_src_in_bund(is, bn, e, iw=f, ic=ic))
    call  clanRcvrIdxRng_inBun(rngB, bn%c(ic), bn)
    f2  = f*2
	w   = e%s%w1 + e%s%dw* (f-1)
    g1x = e%s%g1 + e%s%dg* (bn%c(ic)%rj1 -1)
	g2x = e%s%g1 + e%s%dg* (bn%c(ic)%rj2 -1)
	
	   p(g1x :g2x: e%s%dg, e%s%zg) =           &
	   p(g1x :g2x: e%s%dg, e%s%zg) +           &
    vdt2(g1x :g2x: e%s%dg, e%s%zg) *           &
	    csge(f2-1, rngB(1):rngB(2))*           &
	cos(csge(f2,   rngB(1):rngB(2))+ w*((tt-1)*e%d%dt) )
  end do
end if
end subroutine
!========== DBG ===========
subroutine inject_sources_DBG(xh,xt, bn, e)
integer,             intent(in)   :: xh,xt
type(bund_type),     intent(in)   :: bn
type(entire_param),  intent(in)   :: e

integer              :: tt, is,  iw,  ic, sx,  sht, flip
real                 :: w
real,allocatable     :: p(:, :, :) !  p: pressure time history, 1 z-layer only

if(rank/=nsize-1)  then
  call MPI_barrier(MPI_COMM_WORLD, ierr); return
end if

allocate(p(xh:xt, 1, e%d%nt))
p = 0.
do tt = 1, e%d%nt
    if(e%o%SINU==0)  then   !  wavelet in t-domain, no clan trick
	  do while(iter_src_in_bund(is, bn, e))
		sx = e%s%s1 + (is-1)*e%s%ds ! p must be big enough to contain all source locations
		sht= tt- bn%dly(is)
		if (sht <= 0)  cycle
		p(sx,1,tt) = e%s%s(sht) * bn%pol(is)
	  end do
	else ! sinusoid
	  do while(iter_src_in_bund(is, bn, e, iw=iw, ic=ic))
		w  = e%s%w1 + (iw-1)*e%s%dw
		sx = e%s%s1 + (is-1)*e%s%ds
		p(sx,1,tt) = &
			 e%s%sw(iw) * bn%pol(is) * cos( w*((tt-1)*e%d%dt) + e%s%phw(iw) )
	  end do
	end if
end do

call write_binfile('snap/injec.bin', p)
deallocate( p )
call MPI_barrier(MPI_COMM_WORLD, ierr)
end subroutine
!.....................................
!subroutine nega_d_dt2_coef(csge, b, e, coef)  ! no longer needed
!! apply -d/dt^2 to csge, then * coef 
!! one method of forming image condition.
!real,                 intent(inout) :: csge(:,:)
!type(bund_type),      intent(in)    :: b
!type(entire_param),   intent(inout) :: e
!real, optional,       intent(in)    :: coef
!integer               :: is, f, f2, ic, ig, rngB(2)
!real                  :: w
!if (e%o%SINU /= 0)    then
!  do while(iter_src_in_bund(is, b, e, iw=f, ic=ic))
!    f2    = f * 2
!	w     = e%s%w1 + (f-1)*e%s%dw
!    call    clanRcvrIdxRng_inBun(rngB, b%c(ic), b)  ! clan's aperture in csg's dim2
!	! magnitue *= w^2
!	if (present(coef))  then
!	  csge( f2-1, rngB(1):rngB(2) ) = csge( f2-1, rngB(1):rngB(2) ) * (coef *w*w)
!	else
!      csge( f2-1, rngB(1):rngB(2) ) = csge( f2-1, rngB(1):rngB(2) ) * (w*w)
!    end if	
!  end do
!else
!  do ig = b%rj1, b%rj2
!    e%s%trctmp = csge(:, ig)
!    call nega_diff2( csge(:, ig), e%s%trctmp )
!	if (present(coef))  then
!	  csge(:, ig) = csge(:, ig) * (coef/e%d%dt**2)
!	else
!	  csge(:, ig) = csge(:, ig) / e%d%dt**2
!	end if
!  end do
!end if
!end subroutine
!.....................................
END MODULE
